"""File to contain tests of the embed.py script."""

from nbed.embed import nbed
from nbed.driver import NbedDriver
import json
import pytest
from pydantic import ValidationError


def test_config_input(nbed_config) -> None:
    """Test nbed"""
    assert isinstance(nbed(nbed_config), NbedDriver)


def test_config_overwrite(nbed_config) -> None:
    driver = nbed(nbed_config, n_active_atoms=1)
    assert driver.config.n_active_atoms == 1


def test_file_input(config_file) -> None:
    assert isinstance(nbed(config_file), NbedDriver)


def test_args_input(config_file) -> None:
    with open(config_file) as f:
        args = json.load(f)
    assert isinstance(nbed(**args), NbedDriver)


def test_none_config_input(nbed_config) -> None:
    args = nbed_config.model_dump()
    assert nbed(config=None, **args).config == nbed_config

    args.pop("geometry")
    with pytest.raises(ValidationError) as excinfo:
        nbed(config=None, **args)


def test_wrong_config_object(nbed_config) -> None:
    args = nbed_config.model_dump()
    assert nbed(config=["a", "list"], **args).config == nbed_config


if __name__ == "__main__":
    pass



"""SPADE Localizer Class."""

import logging

import numpy as np
from pyscf import lib
from scipy import linalg

from .base import OccupiedLocalizer

logger = logging.getLogger(__name__)


class SPADELocalizer(OccupiedLocalizer):
    """Object used to localise molecular orbitals (MOs) using SPADE Localization.

    Running localization returns active and environment systems.

    Args:
        global_scf (scf.StreamObject): PySCF method object.
        n_active_atoms (int): Number of active atoms

    Attributes:
        c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
        c_enviro (np.array): C matrix of localized occupied ennironment MOs
        c_loc_occ_and_virt (np.array): Full localized C_matrix (occpuied and virtual)
        dm_active (np.array): active system density matrix
        dm_enviro (np.array): environment system density matrix
        active_MO_inds (np.array): 1D array of active occupied MO indices
        enviro_MO_inds (np.array): 1D array of environment occupied MO indices
        _c_loc_occ (np.array): C matrix of localized occupied MOs

    Methods:
        run: Main function to run localization.
    """

    def __init__(
        self,
        global_scf: lib.StreamObject,
        n_active_atoms: int,
        max_shells: int = 4,
        n_mo_overwrite: tuple[int | None, int | None] | None = None,
    ):
        """Initialize SPADE Localizer object."""
        self.max_shells = max_shells
        self.n_mo_overwrite = (None, None) if n_mo_overwrite is None else n_mo_overwrite
        self.shells = None
        self.singular_values = None
        self.enviro_selection_condition = None

        super().__init__(
            global_scf,
            n_active_atoms,
        )

    def _localize(
        self,
    ) -> tuple[tuple, tuple | None]:
        """Localise orbitals using SPADE.

        Returns:
            active_MO_inds (np.array): 1D array of active occupied MO indices
            enviro_MO_inds (np.array): 1D array of environment occupied MO indices
            c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
            c_enviro (np.array): C matrix of localized occupied ennironment MOs
            c_loc_occ (np.array): full C matrix of localized occupied MOs
        """
        if self.spinless:
            logger.debug("Running SPADE for only one spin.")
            alpha = self._localize_spin(
                self._global_scf.mo_coeff,
                self._global_scf.mo_occ,
                self.n_mo_overwrite[0],
            )
            beta = None
        else:
            alpha = self._localize_spin(
                self._global_scf.mo_coeff[0],
                self._global_scf.mo_occ[0],
                self.n_mo_overwrite[0],
            )
            beta = self._localize_spin(
                self._global_scf.mo_coeff[1],
                self._global_scf.mo_occ[1],
                self.n_mo_overwrite[1],
            )

        return (alpha, beta)

    def _localize_spin(
        self,
        c_matrix: np.ndarray,
        occupancy: np.ndarray,
        n_mo_overwrite: int | None = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Localize orbitals of one spin using SPADE.

        Args:
            c_matrix (np.ndarray): Unlocalized C matrix of occupied orbitals.
            occupancy (np.ndarray): Occupancy of orbitals.
            n_mo_overwrite (int): Number of molecular orbitals to use in active region. Overwrite SVD based value.

        Returns:
            np.ndarray: Localized C matrix of occupied orbitals.
        """
        logger.debug("Localising with SPADE.")

        # We want the same partition for each spin.
        # It wouldn't make sense to have different spin states be localized differently.

        n_occupied_orbitals = np.count_nonzero(occupancy)
        occupied_orbitals = c_matrix[:, :n_occupied_orbitals]
        logger.debug(f"{occupancy=}")
        logger.debug(f"{n_occupied_orbitals} occupied AOs.")

        n_act_aos = self._global_scf.mol.aoslice_by_atom()[self._n_active_atoms - 1][-1]
        logger.debug(f"{n_act_aos} active AOs.")

        ao_overlap = self._global_scf.get_ovlp()

        # Orbital rotation and partition into subsystems A and B
        # rotation_matrix, sigma = embed.orbital_rotation(occupied_orbitals,
        #    n_act_aos, ao_overlap)

        rotated_orbitals = (
            linalg.fractional_matrix_power(ao_overlap, 0.5) @ occupied_orbitals
        )
        _, sigma, right_vectors = linalg.svd(rotated_orbitals[:n_act_aos, :])

        logger.debug(f"Singular Values: {sigma}")

        # n_act_mos, n_env_mos = embed.orbital_partition(sigma)
        # Prevents an error with argmax
        if len(sigma) == 1:
            n_act_mos = 1
        elif n_mo_overwrite is not None and len(sigma) >= n_mo_overwrite:
            logger.debug(f"Enforcing use of {n_mo_overwrite} MOs")
            n_act_mos: int = n_mo_overwrite
        else:
            value_diffs = sigma[:-1] - sigma[1:]
            n_act_mos: int = np.argmax(value_diffs) + 1

        n_env_mos = n_occupied_orbitals - n_act_mos
        logger.debug(f"{n_act_mos} active MOs.")
        logger.debug(f"{n_env_mos} environment MOs.")

        # get active and enviro indices
        active_MO_inds = np.arange(n_act_mos)
        enviro_MO_inds = np.arange(n_act_mos, n_act_mos + n_env_mos)

        # Defining active and environment orbitals and density
        c_active = occupied_orbitals @ right_vectors.T[:, :n_act_mos]
        c_enviro = occupied_orbitals @ right_vectors.T[:, n_act_mos:]
        c_loc_occ = occupied_orbitals @ right_vectors.T

        # storing condition used to select env system
        if self.enviro_selection_condition is None:
            self.enviro_selection_condition = (sigma, np.zeros(len(sigma)))
        else:
            self.enviro_selection_condition = (
                self.enviro_selection_condition[0],
                sigma,
            )

        return (active_MO_inds, enviro_MO_inds, c_active, c_enviro, c_loc_occ)



"""Init for Occupied Localizer classes."""

from .base import OccupiedLocalizer
from .pyscf import BOYSLocalizer, IBOLocalizer, PMLocalizer
from .spade import SPADELocalizer

__all__ = [
    "BOYSLocalizer",
    "IBOLocalizer",
    "PMLocalizer",
    "SPADELocalizer",
    "OccupiedLocalizer",
]



"""Tests for the HamiltonianBuilder class."""

from logging import getLogger

import numpy as np
import pytest
import scipy as sp
from openfermion import count_qubits, get_sparse_operator
from openfermion.ops import InteractionOperator
from openfermion.transforms.opconversions import jordan_wigner
from pyscf.fci import FCI
from pyscf.gto import Mole
from pyscf.scf import RHF, UHF

from nbed.ham_builder import HamiltonianBuilder, reduce_virtuals

logger = getLogger(__name__)


@pytest.fixture
def uncharged_mol(water_filepath) -> dict:
    mol_args = {
        "atom": str(water_filepath),
        "n_active_atoms": 1,
        "basis": "STO-3G",
        "unit": "angstrom",
    }
    return Mole(**mol_args, charge=0, spin=0).build()


@pytest.fixture
def restricted_scf(uncharged_mol):
    rhf = RHF(uncharged_mol)
    rhf.kernel()
    return rhf


@pytest.fixture
def unrestricted_scf(uncharged_mol):
    uhf = UHF(uncharged_mol)
    uhf.kernel()
    return uhf


@pytest.fixture
def rbuilder(restricted_scf):
    return HamiltonianBuilder(restricted_scf, 0).build()


@pytest.fixture
def ubuilder(unrestricted_scf):
    return HamiltonianBuilder(unrestricted_scf, 0).build()


def test_restricted_groundstate(restricted_scf, rbuilder) -> None:
    """Use the full system to check that output hamiltonian diagonalises to fci value for a restricted calculation."""
    e_fci = FCI(restricted_scf).kernel()[0] - restricted_scf.energy_nuc()

    logger.info(f"FCI energy of unrestricted driver test: {e_fci}")

    const, ones, twos = rbuilder
    intop = InteractionOperator(const, ones, twos)
    qham = jordan_wigner(intop)

    assert count_qubits(qham) == 14
    diag, _ = sp.sparse.linalg.eigsh(get_sparse_operator(qham), k=1, which="SA")
    logger.info(f"Ground state via diagonalisation: {diag}")
    assert np.isclose(e_fci, diag)


def test_unrestricted_groundstate(unrestricted_scf, ubuilder) -> None:
    """Use the full system to check that output hamiltonian diagonalises to fci value for a restricted calculation."""
    e_fci = FCI(unrestricted_scf).kernel()[0] - unrestricted_scf.energy_nuc()

    logger.info(f"FCI energy of unrestricted driver test: {e_fci}")

    const, ones, twos = ubuilder
    intop = InteractionOperator(const, ones, twos)
    qham = jordan_wigner(intop)

    assert count_qubits(qham) == 14
    diag, _ = sp.sparse.linalg.eigsh(get_sparse_operator(qham), k=1, which="SA")
    logger.info(f"Ground state via diagonalisation: {diag}")
    assert np.isclose(e_fci, diag)


@pytest.fixture
def charged_mol(water_filepath) -> Mole:
    mol_args = {
        "atom": str(water_filepath),
        "n_active_atoms": 1,
        "basis": "STO-3G",
        "unit": "angstrom",
    }
    return Mole(**mol_args, charge=1, spin=1).build()


@pytest.fixture
def charged_scf(charged_mol):
    rhf = UHF(charged_mol)
    rhf.kernel()
    return rhf


def test_unrestricted_charged_groundstate(charged_scf) -> None:
    """Check the output hamiltonian diagonalises to fci value for an unrestricted calculation with spin and charge."""
    e_fci = FCI(charged_scf).kernel()[0] - charged_scf.energy_nuc()

    logger.info(f"FCI energy of unrestricted driver test: {e_fci}")

    builder = HamiltonianBuilder(charged_scf)
    const, ones, twos = builder.build()
    intop = InteractionOperator(const, ones, twos)
    qham = jordan_wigner(intop)

    diag, _ = sp.sparse.linalg.eigsh(get_sparse_operator(qham), k=2, which="SA")

    logger.info(f"Ground state via diagonalisation: {diag}")
    # Ground state for this charge is 2nd eigenstate
    assert np.isclose(e_fci, diag[1])


def test_reduce_virtuals(restricted_scf, unrestricted_scf):
    reduced_restricted = reduce_virtuals(restricted_scf, 1)
    reduced_unrestricted = reduce_virtuals(unrestricted_scf, 1)

    assert (
        reduced_restricted.mo_coeff.shape[-1]
        == reduced_unrestricted.mo_coeff.shape[-1]
        == 6
    )
    assert np.all(
        reduced_restricted.mo_occ == np.sum(reduced_unrestricted.mo_occ, axis=0)
    )

    with pytest.raises(ValueError) as excinfo:
        reduce_virtuals(restricted_scf, 7)

    assert "more than exist" in str(excinfo)

    assert np.all(
        restricted_scf.mo_coeff == reduce_virtuals(restricted_scf, 0).mo_coeff
    )
    assert np.all(
        restricted_scf.mo_coeff == reduce_virtuals(restricted_scf, 0).mo_coeff
    )



"""Shared fixtures for tests."""

from pathlib import Path

import pytest
from pyscf import gto, scf, dft
from pyscf.lib import StreamObject

from nbed.driver import NbedDriver
from nbed.config import NbedConfig


@pytest.fixture(scope="module")
def config_file() -> Path:
    return Path("tests/test_config.json").absolute()


@pytest.fixture(scope="module")
def water_filepath() -> Path:
    return Path("tests/molecules/water.xyz").absolute()


@pytest.fixture(scope="module")
def pfoa_filepath() -> Path:
    return Path("tests/molecules/pfoa.xyz").absolute()


@pytest.fixture(scope="module")
def water_molecule(water_filepath) -> gto.Mole:
    mol_args = {
        "atom": str(water_filepath),
        "n_active_atoms": 1,
        "basis": "STO-3G",
        "unit": "angstrom",
    }
    return gto.Mole(**mol_args, charge=0, spin=0).build()


@pytest.fixture(scope="module")
def water_rhf(water_molecule) -> StreamObject:
    rhf = scf.rhf.RHF(water_molecule)
    rhf.kernel()
    return rhf


@pytest.fixture(scope="module")
def water_uhf(water_molecule) -> StreamObject:
    uhf = scf.uhf.UHF(water_molecule)
    uhf.kernel()
    return uhf


@pytest.fixture(scope="module")
def water_rks(water_molecule) -> StreamObject:
    rks = dft.RKS(water_molecule)
    rks.kernel()
    return rks


@pytest.fixture(scope="module")
def water_uks(water_molecule) -> StreamObject:
    uks = dft.UKS(water_molecule)
    uks.kernel()
    return uks


@pytest.fixture(scope="module")
def nbed_args(water_filepath) -> dict:
    args = {
        "geometry": water_filepath,
        "n_active_atoms": 1,
        "basis": "STO-3G",
        "xc_functional": "b3lyp",
        "projector": "mu",
        "localization": "spade",
        "convergence": 1e-06,
        "charge": 0,
        "spin": 0,
        "symmetry": False,
        "mu_level_shift": 1000000.0,
        "run_ccsd_emb": True,
        "run_fci_emb": True,
        "run_virtual_localization": False,
        "n_mo_overwrite": (None, None),
        "run_dft_in_dft": False,
        "max_ram_memory": 4000,
        "occupied_threshold": 0.95,
        "virtual_threshold": 0.95,
        "max_shells": 4,
        "init_huzinaga_rhf_with_mu": True,
        "max_hf_cycles": 50,
        "max_dft_cycles": 50,
        "force_unrestricted": False,
        "mm_coords": None,
        "mm_charges": None,
        "mm_radii": None,
    }
    return args


@pytest.fixture(scope="module")
def nbed_config(nbed_args) -> NbedConfig:
    return NbedConfig(**nbed_args)


@pytest.fixture(scope="module")
def spinless_driver():
    water_xyz_raw = (
        "3\n \nH\t0.2774\t0.8929\t0.2544\nO\t0\t0\t0\nH\t0.6068\t-0.2383\t-0.7169"
    )
    args = {
        "geometry": water_xyz_raw,
        "n_active_atoms": 2,
        "basis": "STO-3G",
        "xc_functional": "b3lyp",
        "projector": "mu",
        "localization": "spade",
        "convergence": 1e-6,
        "run_ccsd_emb": False,
        "run_fci_emb": False,
    }

    config = NbedConfig(**args)
    driver = NbedDriver(config)
    driver.embed()
    return driver


@pytest.fixture(scope="module")
def unrestricted_driver():
    water_xyz_raw = (
        "3\n \nH\t0.2774\t0.8929\t0.2544\nO\t0\t0\t0\nH\t0.6068\t-0.2383\t-0.7169"
    )
    args = {
        "geometry": water_xyz_raw,
        "n_active_atoms": 2,
        "basis": "STO-3G",
        "xc_functional": "b3lyp",
        "projector": "mu",
        "localization": "spade",
        "convergence": 1e-6,
        "run_ccsd_emb": False,
        "run_fci_emb": False,
        "force_unrestricted": True,
    }
    config = NbedConfig(**args)
    driver = NbedDriver(config)
    driver.embed()
    return driver



"""Helper functions for the package."""

import argparse
import json
import logging
import logging.config
import os
from pathlib import Path
from typing import Optional

from openfermion.chem.pubchem import geometry_from_pubchem
from pydantic import ValidationError

from nbed.config import NbedConfig

logger = logging.getLogger(__name__)


def setup_logs() -> None:
    """Initialise logging."""
    config_dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {"format": "%(asctime)s: %(name)s: %(levelname)s: %(message)s"},
        },
        "handlers": {
            "file_handler": {
                "class": "logging.FileHandler",
                "level": "DEBUG",
                "formatter": "standard",
                "filename": ".nbed.log",
                "mode": "w",
                "encoding": "utf-8",
            },
            "stream_handler": {
                "class": "logging.StreamHandler",
                "level": "WARNING",
                "formatter": "standard",
            },
        },
        "loggers": {
            "": {"handlers": ["file_handler", "stream_handler"], "level": "DEBUG"}
        },
    }

    logging.config.dictConfig(config_dict)
    logger = logging.getLogger(__name__)
    logger.debug("Logging initialised.")


def parse():
    """Parse arguments from command line interface."""
    logger.debug("Adding CLI arguments.")
    parser = argparse.ArgumentParser(description="Output embedded Qubit Hamiltonian.")
    parser.add_argument(
        "--config",
        required=True,
        type=str,
        help="Path to a config file. Overwrites other arguments.",
    )
    logger.debug("Parsing CLI arguments.")
    args = parser.parse_args()

    logger.debug("Reading config file.")
    filepath = Path(args.config).absolute()
    with open(filepath) as f:
        config_data = json.load(f)
    logger.debug(f"Input data:\n{config_data=}")

    try:
        config = NbedConfig(config_data)
    except ValidationError as e:
        logger.error("Could not validate input data against NbedConfig model.")
        logger.error(e)

    return config


def pubchem_mol_geometry(molecule_name) -> dict:
    """Wrapper of Openfermion function to extract geometry using the molecule's name from the PubChem.

    Returns a dictionary of atomic type and xyz location, each indexed by dictionary key.

    Args:
        molecule_name (str): Name of molecule to search on pubchem
    Returns:
        struct_dict (dict): Keys index atoms and values contain Tuple of ('atom_id', (x_loc, y_loc, z_loc)

    Example:
    output = pubchem_mol_geometry('H2O')
    print(output)

    >> { 0: ('O', (0, 0, 0)),
         1: ('H', (0.2774, 0.8929, 0.2544)),
         2: ('H', (0.6068, -0.2383, -0.7169))
         }

    """
    geometry_pubchem = geometry_from_pubchem(molecule_name, structure="3d")

    if geometry_pubchem is None:
        raise ValueError(
            f"""Could not find geometry of {molecule_name} on PubChem...
                                 make sure molecule input is a correct path to an xyz file or real molecule
                                """
        )

    struct_dict = {}
    for ind, atom_xyz in enumerate(geometry_pubchem):
        struct_dict[ind] = atom_xyz
    return struct_dict


def build_ordered_xyz_string(struct_dict: dict, active_atom_inds: list) -> str:
    """Get raw xyz string of molecular geometry.

    This function orders the atoms in struct_dict according to the ordering given in atom_ordering_by_inds list.

    Args:
        struct_dict (dict): Dictionary of indexed atoms and Cartesian coordinates (x,y,z)
        active_atom_inds (list): list of indices to be considered active. This will put these atoms to the top of the xyz file.
                                 Note indices are chosen from the struct_dict.

    Returns:
        xyz_string (str): raw xyz string of molecular geometry (atoms ordered by atom_ordering_by_inds list)

    Example:
        input_struct_dict = { 0: ('O', (0, 0, 0)),
                              1: ('H', (0.2774, 0.8929, 0.2544)),
                              2: ('H', (0.6068, -0.2383, -0.7169))
                            }

        xyz_string = ordered_xyz_string('water', input_struct_dict, [1,0,2])
        print(xyz_string)

         >> 3

            H	0.2774	0.8929	0.2544
            O	0	0	0
            H	0.6068	-0.2383	-0.7169

    """
    if not set(active_atom_inds).issubset(set(list(struct_dict.keys()))):
        raise ValueError(
            "active atom indices not subset of indices in structural dict "
        )

    ordering = (
        *active_atom_inds,
        *[ind for ind in struct_dict.keys() if ind not in active_atom_inds],
    )

    n_atoms = len(struct_dict)
    xyz_file = f"{n_atoms}"
    xyz_file += "\n \n"
    for atom_ind in ordering:
        atom, xyz = struct_dict[atom_ind]
        xyz_file += f"{atom}\t{xyz[0]}\t{xyz[1]}\t{xyz[2]}\n"

    return xyz_file


def save_ordered_xyz_file(
    file_name: str,
    struct_dict: dict,
    active_atom_inds: list,
    save_location: Optional[Path] = None,
) -> Path:
    """Saves .xyz file in a molecular_structures directory.

    This function orders the atoms in struct_dict according to the ordering
    given in atom_ordering_by_inds list. The file is then saved.
    The location of this director is either at save_location, or if not defined then in current working dir.
    Function returns the path to xyz file.

    Args:
        file_name (str): Name of xyz file
        struct_dict (dict): Dictionary of indexed atoms and Cartesian coordinates (x,y,z)
        active_atom_inds (list): list of indices to be considered active. This will put these atoms to the top of the xyz file.
                                 Note indices are chosen from the struct_dict.
        save_location (Path): Path of where to save xyz file. If not defined then current working dir used.

    Returns:
        xyz_file_path (Path): Path to xyz file

    Example:
    input_struct_dict = { 0: ('O', (0, 0, 0)),
                            1: ('H', (0.2774, 0.8929, 0.2544)),
                            2: ('H', (0.6068, -0.2383, -0.7169))
                        }

    path = save_ordered_xyz_file('water', input_struct_dict, [1,0,2])
    print(path)
    >> ../molecular_structures/water.xyz

    with open(path,'r') as infile:
        xyz_string = infile.read()
    print(xyz_string)

        >> 3

        H	0.2774	0.8929	0.2544
        O	0	0	0
        H	0.6068	-0.2383	-0.7169

    """
    xyz_string = build_ordered_xyz_string(struct_dict, active_atom_inds)

    if save_location is None:
        save_location = Path(os.getcwd())

    output_dir = os.path.join(str(save_location), "molecular_structures")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    xyz_file_path = os.path.join(output_dir, f"{file_name}.xyz")

    with open(xyz_file_path, "w") as outfile:
        outfile.write(xyz_string)

    return Path(xyz_file_path)



"""Functions from PySCF that need to be tweeked to allow our hack of adding Vemb to Hcore."""

import logging

import numpy as np
from pyscf import ao2mo

logger = logging.getLogger(__name__)


def energy_elec(mf, dm=None, h1e=None, vhf=None) -> tuple[float, float]:
    """Electronic energy of Unrestricted Hartree-Fock.

    Note this function has side effects which cause mf.scf_summary updated.

    Args:
        mf (pyscf.scf.hf.HF): Hartree-Fock object
        dm (np.ndarray): Density matrix
        h1e (np.ndarray): Core Hamiltonian
        vhf (np.ndarray): 2-electron contribution to effective potential

    Returns:
        e_elec (np.ndarray): Hartree-Fock electronic energy
        e_coul (np.ndarray): 2-electron contribution to electronic energy
    """
    logger.debug("Running spin-aware energy_elec")
    if dm is None:
        dm = mf.make_rdm1()
    if h1e is None:
        h1e = mf.get_hcore()
    if isinstance(dm, np.ndarray) and dm.ndim == 2:
        dm = np.array((dm * 0.5, dm * 0.5))
    if vhf is None:
        vhf = mf.get_veff(mf.mol, dm)
    logger.debug(f"{h1e.shape=}")
    logger.debug(f"{dm.shape=}")
    logger.debug(f"{vhf.shape=}")
    e1 = np.einsum("ij,ji->", h1e[0], dm[0])
    e1 += np.einsum("ij,ji->", h1e[1], dm[1])
    e_coul = (
        np.einsum("ij,ji->", vhf[0], dm[0]) + np.einsum("ij,ji->", vhf[1], dm[1])
    ) * 0.5
    e_elec = (e1 + e_coul).real
    mf.scf_summary["e1"] = e1.real
    mf.scf_summary["e2"] = e_coul.real
    return e_elec, e_coul


def _absorb_h1e(h1e, eri, norb, nelec, fac=1):
    if not isinstance(nelec, (int, np.number)):
        nelec = sum(nelec)
    h1e_a, h1e_b = h1e
    print("h1e_a", h1e_a.shape)
    print("h1e_b", h1e_b.shape)
    h1e_a = np.einsum("jik->jk", h1e_a)
    h1e_b = np.einsum("jik->jk", h1e_b)
    print("h1e_a", h1e_a.shape)
    print("h1e_b", h1e_b.shape)
    h2e_aa = ao2mo.restore(1, eri[0], norb).copy()
    h2e_ab = ao2mo.restore(1, eri[1], norb).copy()
    h2e_bb = ao2mo.restore(1, eri[2], norb).copy()
    print("x1", h2e_aa.shape)
    print("y2", h2e_bb.shape)
    x = np.einsum("jiik->jk", h2e_aa) * 0.5
    y = np.einsum("jiik->jk", h2e_bb) * 0.5
    print("x", x.shape)
    print("y", y.shape)
    f1e_a = h1e_a - np.einsum("jiik->jk", h2e_aa) * 0.5
    f1e_b = h1e_b - np.einsum("jiik->jk", h2e_bb) * 0.5
    f1e_a *= 1.0 / (nelec + 1e-100)
    f1e_b *= 1.0 / (nelec + 1e-100)
    for k in range(norb):
        h2e_aa[:, :, k, k] += f1e_a
        h2e_aa[k, k, :, :] += f1e_a
        h2e_ab[:, :, k, k] += f1e_a
        h2e_ab[k, k, :, :] += f1e_b
        h2e_bb[:, :, k, k] += f1e_b
        h2e_bb[k, k, :, :] += f1e_b
    return (
        ao2mo.restore(4, h2e_aa, norb) * fac,
        ao2mo.restore(4, h2e_ab, norb) * fac,
        ao2mo.restore(4, h2e_bb, norb) * fac,
    )



"""Tests for localization functions."""

import numpy as np
import pytest
from pyscf import gto, scf

from nbed.localizers.occupied import OccupiedLocalizer, PMLocalizer, SPADELocalizer
from nbed.localizers.virtual import ConcentricLocalizer
from nbed.localizers.ace import ACELocalizer

xc_functional = "b3lyp"
convergence = 1e-6
pyscf_print_level = 1
max_ram_memory = 4_000
n_active_atoms = 1
occ_cutoff = 0.95
virt_cutoff = 0.95
run_virtual_localization = False


@pytest.fixture
def molecule(water_filepath) -> gto.Mole:
    return gto.Mole(
        atom=str(water_filepath),
        basis="6-31g",
        charge=0,
    ).build()


@pytest.fixture
def global_rks(molecule) -> scf.RKS:
    global_rks = scf.RKS(molecule)
    global_rks.conv_tol = convergence
    global_rks.xc = xc_functional
    global_rks.max_memory = max_ram_memory
    global_rks.verbose = pyscf_print_level
    global_rks.kernel()
    return global_rks


@pytest.fixture
def global_uks(molecule) -> scf.UKS:
    global_uks = scf.UKS(molecule)
    global_uks.conv_tol = convergence
    global_uks.xc = xc_functional
    global_uks.max_memory = max_ram_memory
    global_uks.verbose = pyscf_print_level
    global_uks.kernel()
    return global_uks


def test_base_localizer(global_rks) -> None:
    """Check the base class can be instantiated."""
    with pytest.raises(TypeError) as excinfo:
        OccupiedLocalizer(global_rks, n_active_atoms=n_active_atoms)

    assert "_localize_spin" in str(excinfo.value)


def test_PM_arguments(global_rks) -> None:
    """Check the internal test of values."""
    with pytest.raises(ValueError):
        PMLocalizer(
            global_rks,
            n_active_atoms=n_active_atoms,
            occ_cutoff=1.1,
            virt_cutoff=virt_cutoff,
        )

    with pytest.raises(ValueError):
        PMLocalizer(
            global_rks,
            n_active_atoms=n_active_atoms,
            occ_cutoff=occ_cutoff,
            virt_cutoff=1.1,
        )

    with pytest.raises(ValueError):
        PMLocalizer(
            global_rks,
            n_active_atoms=n_active_atoms,
            occ_cutoff=-0.1,
            virt_cutoff=virt_cutoff,
        )

    with pytest.raises(ValueError):
        PMLocalizer(
            global_rks,
            n_active_atoms=n_active_atoms,
            occ_cutoff=occ_cutoff,
            virt_cutoff=-0.1,
        )


def test_PM_check_values(global_rks, global_uks) -> None:
    """Check the internal test of values."""
    for ks in [global_rks, global_uks]:
        PMLocalizer(
            ks,
            n_active_atoms=n_active_atoms,
            occ_cutoff=occ_cutoff,
            virt_cutoff=virt_cutoff,
        ).run(check_values=True)


def test_SPADE_check_values(global_rks, global_uks) -> None:
    """Check the internal test of values."""
    for ks in [global_rks, global_uks]:
        SPADELocalizer(
            ks,
            n_active_atoms=n_active_atoms,
        ).run(check_values=True)


def test_PM_mo_indices(global_rks, global_uks) -> None:
    restricted_loc_system = PMLocalizer(
        global_rks,
        n_active_atoms=n_active_atoms,
        occ_cutoff=occ_cutoff,
        virt_cutoff=virt_cutoff,
    )
    assert restricted_loc_system.beta_active_MO_inds is None
    assert restricted_loc_system.beta_enviro_MO_inds is None
    assert restricted_loc_system.beta_c_active is None
    assert restricted_loc_system.beta_c_enviro is None
    assert restricted_loc_system._beta_c_loc_occ is None

    unrestricted_loc_system = PMLocalizer(
        global_uks,
        n_active_atoms=n_active_atoms,
        occ_cutoff=occ_cutoff,
        virt_cutoff=virt_cutoff,
    )
    assert np.all(
        restricted_loc_system.active_MO_inds == unrestricted_loc_system.active_MO_inds
    )
    assert np.all(
        restricted_loc_system.enviro_MO_inds == unrestricted_loc_system.enviro_MO_inds
    )
    assert np.all(
        unrestricted_loc_system.active_MO_inds
        == unrestricted_loc_system.beta_active_MO_inds
    )
    assert np.all(
        unrestricted_loc_system.enviro_MO_inds
        == unrestricted_loc_system.beta_enviro_MO_inds
    )


def test_SPADE_mo_indices(global_rks, global_uks) -> None:
    restricted_loc_system = SPADELocalizer(
        global_rks,
        n_active_atoms=n_active_atoms,
    )
    assert restricted_loc_system.beta_active_MO_inds is None
    assert restricted_loc_system.beta_enviro_MO_inds is None
    assert restricted_loc_system.beta_c_active is None
    assert restricted_loc_system.beta_c_enviro is None
    assert restricted_loc_system._beta_c_loc_occ is None

    unrestricted_loc_system = SPADELocalizer(
        global_uks,
        n_active_atoms=n_active_atoms,
    )
    assert np.all(
        restricted_loc_system.active_MO_inds == unrestricted_loc_system.active_MO_inds
    )
    assert np.all(
        restricted_loc_system.enviro_MO_inds == unrestricted_loc_system.enviro_MO_inds
    )
    assert np.all(
        unrestricted_loc_system.active_MO_inds
        == unrestricted_loc_system.beta_active_MO_inds
    )
    assert np.all(
        unrestricted_loc_system.enviro_MO_inds
        == unrestricted_loc_system.beta_enviro_MO_inds
    )


def test_PMLocalizer_local_basis_transform(global_rks) -> None:
    """Check change of basis operator (from canonical to localized) is correct"""
    # run Localizer
    loc_system = PMLocalizer(
        global_rks,
        n_active_atoms=n_active_atoms,
        occ_cutoff=occ_cutoff,
        virt_cutoff=virt_cutoff,
    )
    dm_full_std = global_rks.make_rdm1()
    dm_active_sys = loc_system.dm_active
    dm_enviro_sys = loc_system.dm_enviro
    # y_active + y_enviro = y_total
    assert np.allclose(dm_full_std, dm_active_sys + dm_enviro_sys)

    n_all_electrons = global_rks.mol.nelectron
    s_ovlp = global_rks.get_ovlp()
    n_active_electrons = np.trace(dm_active_sys @ s_ovlp)
    n_enviro_electrons = np.trace(dm_enviro_sys @ s_ovlp)

    # check number of electrons is still the same after orbitals have been localized (change of basis)
    assert np.isclose(n_all_electrons, n_active_electrons + n_enviro_electrons)


def test_spade_spins_match(global_rks, global_uks) -> None:
    """Check that localization of restricted and unrestricted match."""
    # define RKS DFT object

    restricted = SPADELocalizer(
        global_rks,
        n_active_atoms=n_active_atoms,
    )

    unrestricted = SPADELocalizer(
        global_uks,
        n_active_atoms=n_active_atoms,
    )

    # assert loc_system.active_MO_inds
    assert restricted.beta_active_MO_inds is None
    assert np.all(unrestricted.active_MO_inds == unrestricted.beta_active_MO_inds)
    assert np.all(restricted.active_MO_inds == unrestricted.active_MO_inds)


def test_cl_shell_numbers(global_rks, global_uks) -> None:
    restricted_occ = SPADELocalizer(
        global_rks,
        n_active_atoms=n_active_atoms,
    )
    restricted_virt = ConcentricLocalizer(
        restricted_occ._global_scf, n_active_atoms=n_active_atoms
    )
    restricted_virt.localize_virtual()

    unrestricted_occ = SPADELocalizer(
        global_uks,
        n_active_atoms=n_active_atoms,
    )
    unrestricted_virt = ConcentricLocalizer(
        unrestricted_occ._global_scf, n_active_atoms=n_active_atoms
    )
    unrestricted_virt.localize_virtual()

    assert restricted_virt.shells == [12, 13]
    assert (
        restricted_virt.shells
        == unrestricted_virt.shells[0]
        == unrestricted_virt.shells[1]
    )


def test_ace_localizer(global_rks, global_uks) -> None:
    restricted = ACELocalizer(
        global_scf_list=[global_rks] * 3, n_active_atoms=n_active_atoms
    ).localize_path()
    unrestricted = ACELocalizer(
        global_scf_list=[global_uks] * 3, n_active_atoms=n_active_atoms
    ).localize_path()

    restricted_spade = SPADELocalizer(
        global_rks,
        n_active_atoms=n_active_atoms,
        n_mo_overwrite=restricted,
    )

    unrestricted_spade = SPADELocalizer(
        global_uks,
        n_active_atoms=n_active_atoms,
        n_mo_overwrite=unrestricted,
    )
    print(restricted_spade.enviro_selection_condition)
    print(unrestricted_spade.enviro_selection_condition)
    assert restricted == unrestricted == (3, 3)
    assert restricted[0] == restricted[1]
    assert unrestricted[0] == unrestricted[1]
    assert np.all(
        restricted[0] - 1
        == np.argmax(
            restricted_spade.enviro_selection_condition[0][:-1]
            - restricted_spade.enviro_selection_condition[0][1:]
        )
    )
    assert np.all(
        unrestricted[0] - 1
        == np.argmax(
            unrestricted_spade.enviro_selection_condition[0][:-1]
            - unrestricted_spade.enviro_selection_condition[0][1:]
        )
    )


if __name__ == "__main__":
    pass



"""Base Virtual Localizer Class."""

from abc import ABC, abstractmethod

from pyscf import gto
from pyscf.lib import StreamObject


class VirtualLocalizer(ABC):
    """Base class for virtual localizers.

    Args:
        embedded_scf (StreamObject): SCF object with occupied orbitals localized.
        n_active_atoms (int): Number of active atoms in the system.

    Attributes:
        embedded_scf (StreamObject): SCF object with occupied orbitals localized.
        n_active_atoms (int): Number of active atoms in the system.
    """

    def __init__(self, embedded_scf: StreamObject):
        """Initialize VirtualLocalizer.

        Args:
            embedded_scf (StreamObject): A pyscf SCF object.
        """
        self.embedded_scf = embedded_scf

    @abstractmethod
    def localize_virtual(self) -> gto.Mole:
        """Localize virtual orbitals.

        Returns:
            gto.Mole: Localized SCF object.
        """
        pass



"""Tests for utility functions."""

import os

import pytest

from nbed.utils import (
    build_ordered_xyz_string,
    save_ordered_xyz_file,
)


@pytest.fixture
def water_struct_dict() -> dict[int, tuple[str, tuple[float, float, float]]]:
    return {
        0: ("O", (0, 0, 0)),
        1: ("H", (0.2774, 0.8929, 0.2544)),
        2: ("H", (0.6068, -0.2383, -0.7169)),
    }


@pytest.fixture
def water_xyz_o_active():
    return "3\n \nO\t0\t0\t0\nH\t0.2774\t0.8929\t0.2544\nH\t0.6068\t-0.2383\t-0.7169\n"


# def test_pubchem_mol_geometry(water_struct_dict) -> None:

#     assert pubchem_mol_geometry("water") == water_struct_dict


def test_build_ordered_xyz_string(water_struct_dict, water_xyz_o_active) -> None:
    assert build_ordered_xyz_string(water_struct_dict, [0]) == water_xyz_o_active

    water_xyz_h_active = (
        "3\n \nH\t0.2774\t0.8929\t0.2544\nH\t0.6068\t-0.2383\t-0.7169\nO\t0\t0\t0\n"
    )

    assert build_ordered_xyz_string(water_struct_dict, [1, 2]) == water_xyz_h_active


def test_save_xyz(water_struct_dict, water_xyz_o_active) -> None:
    file_path = save_ordered_xyz_file(
        file_name="water_test", struct_dict=water_struct_dict, active_atom_inds=[0]
    )
    f = open(file_path)

    try:
        assert f.read() == water_xyz_o_active
    except AssertionError:
        raise AssertionError("File not saved correctly")
    finally:
        f.close()
        os.remove(file_path)



"""Base Localizer Class."""

import logging
from abc import ABC, abstractmethod
from typing import Tuple, Union

import numpy as np
from pyscf.lib import StreamObject

from ...exceptions import NbedLocalizerError

logger = logging.getLogger(__name__)


class OccupiedLocalizer(ABC):
    """Object used to localise molecular orbitals (MOs) using different localization schemes.

    Running localization returns active and environment systems.

    Note:
    The major improvement of IBOs over PM orbitals is that they are based on IAO charges instead of the erratic
    Mulliken charges. As a result, IBOs are always well-defined.  (Ref: J. Chem. Theory Comput. 2013, 9, 4834−4843)

    Args:
        global_scf (gto.Mole): PySCF molecule object
        n_active_atoms (int): Number of active atoms

    Attributes:
        c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
        c_enviro (np.array): C matrix of localized occupied ennironment MOs
        c_loc_occ_and_virt (np.array): Full localized C_matrix (occpuied and virtual)
        dm_active (np.array): active system density matrix
        dm_enviro (np.array): environment system density matrix
        active_MO_inds (np.array): 1D array of active occupied MO indices
        enviro_MO_inds (np.array): 1D array of environment occupied MO indices
        _c_loc_occ (np.array): C matrix of localized occupied MOs

    Methods:
        run: Main function to run localization.
    """

    def __init__(
        self,
        global_scf: StreamObject,
        n_active_atoms: int,
    ):
        """Initialise class."""
        logger.debug("Initialising LocalizerEnum.")
        if global_scf.mo_coeff is None:
            logger.debug("SCF method not initialised, running now...")
            global_scf.run()
            logger.debug("SCF method initialised.")

        self._global_scf = global_scf
        self._n_active_atoms = n_active_atoms
        if global_scf.mo_coeff.ndim == 2:
            self.spinless = True
        else:
            self.spinless = False
        logger.debug(f"Global scf: {type(global_scf)}")

        # Run the localization procedure
        self.run()

    def _localize(
        self,
    ) -> tuple[Tuple, Union[Tuple, None]]:
        """Localise orbitals using SPADE.

        Returns:
            active_MO_inds (np.array): 1D array of active occupied MO indices
            enviro_MO_inds (np.array): 1D array of environment occupied MO indices
            c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
            c_enviro (np.array): C matrix of localized occupied ennironment MOs
            c_loc_occ (np.array): full C matrix of localized occupied MOs
        """
        if self.spinless:
            alpha = self._localize_spin(
                self._global_scf.mo_coeff, self._global_scf.mo_occ
            )
            beta = None
        else:
            alpha = self._localize_spin(
                self._global_scf.mo_coeff[0], self._global_scf.mo_occ[0]
            )
            beta = self._localize_spin(
                self._global_scf.mo_coeff[1], self._global_scf.mo_occ[1]
            )

        return (alpha, beta)

    @abstractmethod
    def _localize_spin(
        self, c_matrix: np.ndarray, occupancy: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Localize orbitals of one spin.

        Args:
            c_matrix (np.ndarray): Unlocalized C matrix of occupied orbitals.
            occupancy (np.ndarray): Occupancy of orbitals.

        Returns:
            np.ndarray: Localized C matrix of occupied orbitals.
        """
        pass

    def _check_values(self) -> None:  # Needs clarification
        """Check that output values make sense.

        - Same number of active and environment orbitals in alpha and beta
        - Total DM is sum of active and environment DM
        - Total number of electrons conserved

        """
        logger.debug("Running localizer sense check.")
        warn_flag = False
        if self.spinless is False:
            logger.debug("Checking spin does not affect localization.")
            active_number_match = (
                self.active_MO_inds.shape == self.beta_active_MO_inds.shape
            )
            logger.debug(f"{active_number_match=}")
            enviro_number_match = (
                self.enviro_MO_inds.shape == self.beta_enviro_MO_inds.shape
            )
            logger.debug(f"{enviro_number_match=}")
            if not active_number_match or not enviro_number_match:
                logger.error("Number of alpha and beta orbitals do not match.")
                logger.debug(
                    f"alpha: {self.active_MO_inds.shape} active, {self.enviro_MO_inds.shape} enviro"
                )
                logger.debug(
                    f"beta: {self.beta_active_MO_inds.shape} active, {self.beta_enviro_MO_inds.shape} enviro"
                )
                warn_flag = True

        # checking denisty matrix parition sums to total
        logger.debug("Checking density matrix partition.")
        dm_localised_full_system = self._c_loc_occ @ self._c_loc_occ.conj().T
        dm_sum = self.dm_active + self.dm_enviro
        if self.spinless:
            # In a restricted system we have two electrons per orbital
            density_match = np.allclose(2 * dm_localised_full_system, dm_sum)
            logger.debug(f"Restricted {density_match=}")
        else:
            beta_dm_localised_full_system = (
                self._beta_c_loc_occ @ self._beta_c_loc_occ.conj().T
            )
            beta_dm_sum = self.beta_dm_active + self.beta_dm_enviro

            # both need to be correct
            alpha_density_match = np.allclose(dm_localised_full_system, dm_sum)
            logger.debug(f"Unrestricted {alpha_density_match=}")
            beta_density_match = np.allclose(beta_dm_localised_full_system, beta_dm_sum)
            logger.debug(f"Unrestricted {beta_density_match=}")
            density_match = alpha_density_match and beta_density_match

        if not density_match:
            logger.error("Density matrix partition does not sum to total.")
            warn_flag = True

        # check number of electrons is still the same after orbitals have been localized (change of basis)
        logger.debug("Checking electron number conserverd.")
        s_ovlp = self._global_scf.get_ovlp()
        n_active_electrons = np.trace(self.dm_active @ s_ovlp)
        n_enviro_electrons = np.trace(self.dm_enviro @ s_ovlp)

        if self.spinless is False:
            n_active_electrons += np.trace(self.beta_dm_active @ s_ovlp)
            n_enviro_electrons += np.trace(self.beta_dm_enviro @ s_ovlp)

        n_all_electrons = self._global_scf.mol.nelectron
        electron_number_match = np.isclose(
            (n_active_electrons + n_enviro_electrons), n_all_electrons
        )
        logger.debug(f"{electron_number_match=}")
        if not electron_number_match:
            logger.error("Number of electrons in localized orbitals is not consistent.")
            logger.debug(f"N total electrons: {n_all_electrons}")
            warn_flag = True

        if warn_flag:
            logger.error("Localizer sense check failed.")
            raise NbedLocalizerError("Localizer sense check failed.\n")
        else:
            logger.debug("Localizer sense check passed.")

    def run(self, check_values: bool = False) -> None:
        """Function that runs localization.

        Args:
            check_values (bool): optional flag to check denisty matrices and electron number after orbital localization
                                 makes sense
        """
        alpha, beta = self._localize()

        (
            self.active_MO_inds,
            self.enviro_MO_inds,
            self.c_active,
            self.c_enviro,
            self._c_loc_occ,
        ) = alpha

        self.dm_active = self.c_active @ self.c_active.T
        self.dm_enviro = self.c_enviro @ self.c_enviro.T

        # For resticted methods
        if beta is None:
            self.dm_active *= 2.0
            self.dm_enviro *= 2.0
            self.beta_active_MO_inds = None
            self.beta_enviro_MO_inds = None
            self.beta_c_active = None
            self.beta_c_enviro = None
            self._beta_c_loc_occ = None
            self.beta_dm_active = None
            self.beta_dm_enviro = None
        else:
            (
                self.beta_active_MO_inds,
                self.beta_enviro_MO_inds,
                self.beta_c_active,
                self.beta_c_enviro,
                self._beta_c_loc_occ,
            ) = beta

            self.beta_dm_active = self.beta_c_active @ self.beta_c_active.T
            self.beta_dm_enviro = self.beta_c_enviro @ self.beta_c_enviro.T

        if check_values is True:
            self._check_values()

        logger.debug("Localization complete.")
        logger.debug("Localized orbitals:")
        logger.debug("Alpha spin")
        logger.debug(f"{self.active_MO_inds=}")
        logger.debug(f"{self.enviro_MO_inds=}")
        logger.debug(f"{self.c_active.shape=}")
        logger.debug(f"{self.c_enviro.shape=}")
        logger.debug(f"{self._c_loc_occ.shape=}")
        logger.debug("Beta spin")
        logger.debug(f"{self.beta_active_MO_inds=}")
        logger.debug(f"{self.beta_enviro_MO_inds=}")
        if beta is not None:
            logger.debug(f"{self.beta_c_enviro.shape=}")
            logger.debug(f"{self.beta_c_active.shape=}")
            logger.debug(f"{self._beta_c_loc_occ.shape=}")



"""Init Localizer classes."""

from .ace import ACELocalizer
from .occupied.base import OccupiedLocalizer
from .occupied.pyscf import BOYSLocalizer, IBOLocalizer, PMLocalizer
from .occupied.spade import SPADELocalizer
from .virtual.concentric import ConcentricLocalizer

__all__ = [
    "BOYSLocalizer",
    "IBOLocalizer",
    "PMLocalizer",
    "SPADELocalizer",
    "ACELocalizer",
    "ConcentricLocalizer",
    "OccupiedLocalizer",
]



"""PySCF Localizer Class."""

import logging
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
from pyscf import lo
from pyscf.lib import StreamObject
from pyscf.lo import vvo

from .base import OccupiedLocalizer

logger = logging.getLogger(__name__)


class PySCFLocalizer(OccupiedLocalizer, ABC):
    """Object used to localise molecular orbitals (MOs) using PySCF localization functions.

    Running localization returns active and environment systems.

    Note:
    The major improvement of IBOs over PM orbitals is that they are based on IAO charges instead of the erratic
    Mulliken charges. As a result, IBOs are always well-defined.  (Ref: J. Chem. Theory Comput. 2013, 9, 4834−4843)


    Args:
        global_scf (gto.Mole): PySCF molecule object
        n_active_atoms (int): Number of active atoms
        occ_cutoff (float): Threshold for selecting occupied active region (only requried if
                                spade localization is NOT used)
        virt_cutoff (float): Threshold for selecting unoccupied (virtual) active region (required for
                                spade approach too!)

    Attributes:
        c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
        c_enviro (np.array): C matrix of localized occupied ennironment MOs
        c_loc_occ_and_virt (np.array): Full localized C_matrix (occpuied and virtual)
        dm_active (np.array): active system density matrix
        dm_enviro (np.array): environment system density matrix
        active_MO_inds (np.array): 1D array of active occupied MO indices
        enviro_MO_inds (np.array): 1D array of environment occupied MO indices
        _c_loc_occ (np.array): C matrix of localized occupied MOs

    Methods:
        run: Main function to run localization.
    """

    def __init__(
        self,
        global_scf: StreamObject,
        n_active_atoms: int,
        occ_cutoff: Optional[float] = 0.95,
        virt_cutoff: Optional[float] = 0.95,
    ):
        """Initialize PySCF Localizer."""
        self.occ_cutoff = self._valid_threshold(occ_cutoff)
        self.virt_cutoff = self._valid_threshold(virt_cutoff)
        super().__init__(
            global_scf,
            n_active_atoms,
        )

    def _valid_threshold(self, threshold: float):
        """Checks if threshold is within 0-1 range (percentage).

        Args:
            threshold (float): input number between 0 and 1 (inclusive)

        Returns:
            threshold (float): input percentage
        """
        if 0.0 <= threshold <= 1.0:
            logger.debug("Localizer threshold valid.")
            return threshold
        else:
            logger.error("Localizer threshold not valid.")
            raise ValueError(f"threshold: {threshold} is not in range [0,1] inclusive")

    @abstractmethod
    def _pyscf_method(self, c_std_occ):
        """Abstract method containing the PySCF method to use.

        Args:
            c_std_occ (np.ndarray): Unlocalized C matrix of occupied orbitals.
        """
        pass

    def _localize_spin(
        self, c_matrix: np.ndarray, occupancy: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Localize orbitals of one spin using PySCF.

        Args:
            c_matrix (np.ndarray): Unlocalized C matrix of occupied orbitals.
            occupancy (np.ndarray): Occupancy of orbitals.

        Returns:
            np.ndarray: Localized C matrix of occupied orbitals.
        """
        n_occupied_orbitals = np.count_nonzero(occupancy)
        c_std_occ = c_matrix[:, :n_occupied_orbitals]

        c_loc_occ = self._pyscf_method(c_std_occ)

        ao_slice_matrix = self._global_scf.mol.aoslice_by_atom()

        # TODO: Check the following:
        # S_ovlp = pyscf_scf.get_ovlp()
        # S_half = sp.linalg.fractional_matrix_power(S_ovlp , 0.5)
        # C_loc_occ_ORTHO = S_half@C_loc_occ_full
        # run numerator_all and denominator_all in ortho basis

        # find indices of AO of active atoms
        ao_active_inds = np.arange(
            ao_slice_matrix[0, 2], ao_slice_matrix[self._n_active_atoms - 1, 3]
        )
        # active AOs coeffs for a given MO j
        numerator_all = np.einsum("ij->j", (c_loc_occ[ao_active_inds, :]) ** 2)

        # all AOs coeffs for a given MO j
        denominator_all = np.einsum("ij->j", c_loc_occ**2)

        mo_active_share = numerator_all / denominator_all

        logger.debug(f"(active_AO^2)/(all_AO^2): {np.around(mo_active_share, 4)}")
        logger.debug(f"threshold for active part: {self.occ_cutoff}")

        active_MO_inds = np.where(mo_active_share > self.occ_cutoff)[0]
        # print(active_MO_inds)

        all_ao_shares_same_bool = np.allclose(
            np.zeros_like(mo_active_share),
            mo_active_share - mo_active_share.sum() / len(mo_active_share),
        )

        if all_ao_shares_same_bool:
            # case for highly symmetric molecules
            # overlap is the same everywhere hence everything goes into env or act part

            # edge case put half and half!
            logger.warning(
                "AO subsystem selection % same everywhere. Splitting half and half"
            )
            print(f"mo_active_share: {mo_active_share}")
            active_MO_inds = np.array(range(0, c_loc_occ.shape[1] // 2), dtype=int)
        elif len(active_MO_inds) == 0:
            # if no active indices, then take largest possible overlap
            mo_active_percentage_inshare = mo_active_share.argsort()[::-1]
            active_MO_inds = mo_active_percentage_inshare[:1]  # take first element
            logger.warning("no active AOs - forcing one to be active")
            print(f"active system %: {mo_active_share[active_MO_inds][0]} \n")

        enviro_MO_inds = np.array(
            [i for i in range(c_loc_occ.shape[1]) if i not in active_MO_inds]
        )

        # define active MO orbs and environment
        #    take MO (columns of C_matrix) that have high dependence from active AOs
        c_active = c_loc_occ[:, active_MO_inds]

        if len(enviro_MO_inds) == 0:
            # case for when no environement
            logger.warning("No environment electronic density")
            c_enviro = np.zeros((c_active.shape[0], 1))
        else:
            c_enviro = c_loc_occ[:, enviro_MO_inds]

        # storing condition used to select env system
        self.enviro_selection_condition = mo_active_share

        logger.debug("PySCF localization complete.")
        return active_MO_inds, enviro_MO_inds, c_active, c_enviro, c_loc_occ

    def _localize_virtual_spin(
        self, c_matrix: np.ndarray, virt_threshold: float
    ) -> np.ndarray:
        """Localise virtual (unoccupied) orbitals using different localization schemes in PySCF.

        Args:
            global_scf (StreamObject): PySCF molecule object
            c_matrix (np.ndarray): Unlocalized C matrix of occupied orbitals.
            virt_threshold (float): Threshold for selecting unoccupied (virtual) active MOs.

        Returns:
            c_virtual_loc (np.array): C matrix of localized virtual MOs (columns define MOs)
        """
        logger.debug("Localizing virtual orbitals.")
        n_occupied_orbitals = np.count_nonzero(self._global_scf.mo_occ == 2)
        c_std_occ = self._global_scf.mo_coeff[:, :n_occupied_orbitals]
        c_std_virt = self._global_scf.mo_coeff[:, self._global_scf.mo_occ < 2]

        c_virtual_loc = vvo.vvo(
            self._global_scf.mol, c_std_occ, c_std_virt, iaos=None, s=None, verbose=None
        )

        ao_slice_matrix = self._global_scf.mol.aoslice_by_atom()

        # TODO: Check the following:
        # S_ovlp = global_scf.get_ovlp()
        # S_half = sp.linalg.fractional_matrix_power(S_ovlp , 0.5)
        # C_loc_occ_ORTHO = S_half@C_loc_occ_full
        # run numerator_all and denominator_all in ortho basis

        # find indices of AO of active atoms
        ao_active_inds = np.arange(
            ao_slice_matrix[0, 2], ao_slice_matrix[self._n_active_atoms - 1, 3]
        )

        # active AOs coeffs for a given MO j
        numerator_all = np.einsum("ij->j", (c_virtual_loc[ao_active_inds, :]) ** 2)
        # all AOs coeffs for a given MO j
        denominator_all = np.einsum("ij->j", c_virtual_loc**2)

        active_percentage_MO = numerator_all / denominator_all

        logger.debug("Virtual orbitals localized.")
        logger.debug(f"(active_AO^2)/(all_AO^2): {np.around(active_percentage_MO,4)}")
        logger.debug(f"threshold for active part: {self._virt_cutoff}")

        # NOT IN USE
        # add constant occupied index
        # active_virtual_MO_inds = (
        #     np.where(active_percentage_MO > self.virt_cutoff)[0] + c_std_occ.shape[1]
        # )
        # enviro_virtual_MO_inds = np.array(
        #     [
        #         i
        #         for i in range(
        #             c_std_occ.shape[1], c_std_occ.shape[1] + c_virtual_loc.shape[1]
        #         )
        #         if i not in active_virtual_MO_inds
        #     ]
        # )

        return c_virtual_loc

    def localize_virtual(self, local_scf: StreamObject) -> StreamObject:
        """Localise virtual (unoccupied) obitals using PySCF method.

        [1] D. Claudino and N. J. Mayhall, "Simple and Efficient Truncation of Virtual
        Spaces in Embedded Wave Functions via Concentric Localization", Journal of Chemical
        Theory and Computation, vol. 15, no. 11, pp. 6085-6096, Nov. 2019,
        doi: 10.1021/ACS.JCTC.9B00682.

        Args:
            local_scf (StreamObject): SCF object with occupied orbitals localized.

        Returns:
            StreamObject: Fully Localized SCF object.
        """
        raise NotImplementedError(
            "Virtual orbital localization not implemented for PySCF methods."
        )
        return local_scf


class PMLocalizer(PySCFLocalizer):
    """Object used to localise molecular orbitals (MOs) using Pipek-Mezey localization.

    Running localization returns active and environment systems.

    Args:
        global_scf (gto.Mole): PySCF molecule object
        n_active_atoms (int): Number of active atoms
        occ_cutoff (float): Threshold for selecting occupied active region (only requried if
                                spade localization is NOT used)
        virt_cutoff (float): Threshold for selecting unoccupied (virtual) active region (required for
                                spade approach too!)

    Attributes:
        c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
        c_enviro (np.array): C matrix of localized occupied ennironment MOs
        c_loc_occ_and_virt (np.array): Full localized C_matrix (occpuied and virtual)
        dm_active (np.array): active system density matrix
        dm_enviro (np.array): environment system density matrix
        active_MO_inds (np.array): 1D array of active occupied MO indices
        enviro_MO_inds (np.array): 1D array of environment occupied MO indices
        _c_loc_occ (np.array): C matrix of localized occupied MOs

    Methods:
        run: Main function to run localization.
    """

    def __init__(
        self,
        global_scf: StreamObject,
        n_active_atoms: int,
        occ_cutoff: Optional[float] = 0.95,
        virt_cutoff: Optional[float] = 0.95,
    ):
        """Initialize Localizer."""
        super().__init__(
            global_scf,
            n_active_atoms,
            occ_cutoff=occ_cutoff,
            virt_cutoff=virt_cutoff,
        )

    def _pyscf_method(self, c_std_occ: np.ndarray) -> np.ndarray:
        """Abstract method containing the PySCF method to use.

        Args:
            c_std_occ (np.ndarray): Unlocalized C matrix of occupied orbitals.
        """
        logger.debug("Using Pipek-Mezey method.")
        # Localise orbitals using Pipek-Mezey localization scheme.
        # This maximizes the sum of orbital-dependent partial charges on the nuclei.

        pipmez = lo.PipekMezey(self._global_scf.mol, c_std_occ)

        # The atomic population projection scheme.
        # 'mulliken', 'meta-lowdin', 'iao', 'becke'
        pipmez.pop_method = "meta-lowdin"

        # run localization
        return pipmez.kernel()


class BOYSLocalizer(PySCFLocalizer):
    """Object used to localise molecular orbitals (MOs) using BOYS localization.

    Running localization returns active and environment systems.

    Args:
        global_scf (gto.Mole): PySCF molecule object
        n_active_atoms (int): Number of active atoms
        localization_method (str): String of orbital localization method (spade, pipekmezey, boys, ibo)
        occ_cutoff (float): Threshold for selecting occupied active region (only requried if
                                spade localization is NOT used)
        virt_cutoff (float): Threshold for selecting unoccupied (virtual) active region (required for
                                spade approach too!)

    Attributes:
        c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
        c_enviro (np.array): C matrix of localized occupied ennironment MOs
        c_loc_occ_and_virt (np.array): Full localized C_matrix (occpuied and virtual)
        dm_active (np.array): active system density matrix
        dm_enviro (np.array): environment system density matrix
        active_MO_inds (np.array): 1D array of active occupied MO indices
        enviro_MO_inds (np.array): 1D array of environment occupied MO indices
        _c_loc_occ (np.array): C matrix of localized occupied MOs

    Methods:
        run: Main function to run localization.
    """

    def __init__(
        self,
        global_scf: StreamObject,
        n_active_atoms: int,
        occ_cutoff: Optional[float] = 0.95,
        virt_cutoff: Optional[float] = 0.95,
    ):
        """Initialize Localizer."""
        super().__init__(
            global_scf,
            n_active_atoms,
            occ_cutoff=occ_cutoff,
            virt_cutoff=virt_cutoff,
        )

    def _pyscf_method(self, c_std_occ: np.ndarray) -> np.ndarray:
        """Abstract method containing the PySCF method to use.

        Args:
            c_std_occ (np.ndarray): Unlocalized C matrix of occupied orbitals.
        """
        logger.debug("Using BOYS method.")
        #  Minimizes the spatial extent of the orbitals by minimizing a certain function.
        boys_SCF = lo.boys.Boys(self._global_scf.mol, c_std_occ)
        return boys_SCF.kernel()


class IBOLocalizer(PySCFLocalizer):
    """Object used to localise molecular orbitals (MOs) using IBO localization.

    Running localization returns active and environment systems.

    Args:
        global_scf (gto.Mole): PySCF molecule object
        n_active_atoms (int): Number of active atoms
        occ_cutoff (float): Threshold for selecting occupied active region (only requried if
                                spade localization is NOT used)
        virt_cutoff (float): Threshold for selecting unoccupied (virtual) active region (required for
                                spade approach too!)

    Attributes:
        c_active (np.array): C matrix of localized occupied active MOs (columns define MOs)
        c_enviro (np.array): C matrix of localized occupied ennironment MOs
        c_loc_occ_and_virt (np.array): Full localized C_matrix (occpuied and virtual)
        dm_active (np.array): active system density matrix
        dm_enviro (np.array): environment system density matrix
        active_MO_inds (np.array): 1D array of active occupied MO indices
        enviro_MO_inds (np.array): 1D array of environment occupied MO indices
        _c_loc_occ (np.array): C matrix of localized occupied MOs

    Methods:
        run: Main function to run localization.
    """

    def __init__(
        self,
        global_scf: StreamObject,
        n_active_atoms: int,
        occ_cutoff: Optional[float] = 0.95,
        virt_cutoff: Optional[float] = 0.95,
    ):
        """Initialise Localizer."""
        super().__init__(
            global_scf,
            n_active_atoms,
            occ_cutoff=occ_cutoff,
            virt_cutoff=virt_cutoff,
        )

    def _pyscf_method(self, c_std_occ: np.ndarray) -> np.ndarray:
        """Abstract method containing the PySCF method to use.

        Args:
            c_std_occ (np.ndarray): Unlocalized C matrix of occupied orbitals.
        """
        logger.debug("Using IBO method.")
        # Intrinsic bonding orbitals.
        iaos = lo.iao.iao(self._global_scf.mol, c_std_occ)
        # Orthogonalize IAO
        iaos = lo.vec_lowdin(iaos, self._global_scf.get_ovlp())
        c_loc_occ = lo.ibo.ibo(
            self._global_scf.mol, c_std_occ, locmethod="IBO", iaos=iaos
        )
        return c_loc_occ



"""Concentric Localization of virtual orbitals."""

import logging

import numpy as np
from pyscf import gto, scf
from pyscf.lib import StreamObject

from .base import VirtualLocalizer

logger = logging.getLogger(__name__)


class ConcentricLocalizer(VirtualLocalizer):
    """Class to localize virtual orbitals using concentric localization.

    Attributes:
        max_shells (int): Maximum number of shells to localize.
        _n_active_atoms (int): Number of active atoms in the system.
        projected_overlap (np.ndarray): Projected overlap matrix.
        overlap_two_basis (np.ndarray): Overlap matrix between two basis sets.
        n_act_proj_aos (int): Number of active projected atomic orbitals.
        shells (list): List of shell sizes.
        singular_values (list): List of singular values from SVD.

    Methods:
        localize_virtual(StreamObject): Localize virtual orbitals using concentric localization.
        _localize_virtual_spin(np.ndarray, np.ndarray, np.ndarray): Run concentric localization for each spin separately.
    """

    def __init__(
        self,
        embedded_scf: StreamObject,
        n_active_atoms: int,
        max_shells: int = 4,
    ):
        """Initialize Concentric Localization object.

        Args:
            embedded_scf (StreamObject): SCF object with occupied orbitals localized.
            n_active_atoms (int): Number of active atoms in the system.
            max_shells (int): Maximum number of shells to localize.
        """
        super().__init__(embedded_scf)
        self._n_active_atoms = n_active_atoms
        self.max_shells = max_shells
        self.projected_overlap = None
        self.overlap_two_basis = None
        self.n_act_proj_aos = None
        self.shells = None
        self.singular_values = None

    def localize_virtual(self) -> StreamObject:
        """Localise virtual (unoccupied) obitals using concentric localization.

        [1] D. Claudino and N. J. Mayhall, "Simple and Efficient Truncation of Virtual
        Spaces in Embedded Wave Functions via Concentric Localization", Journal of Chemical
        Theory and Computation, vol. 15, no. 11, pp. 6085-6096, Nov. 2019,
        doi: 10.1021/ACS.JCTC.9B00682.

        Returns:
            StreamObject: Fully Localized SCF object.
        """
        logger.debug("Localising virtual orbital spin with concentric localization.")

        logger.debug("Creating projected molecule object.")
        embedded_scf = self.embedded_scf

        projected_mol = gto.mole.Mole()
        projected_mol.atom = embedded_scf.mol.atom
        projected_mol.basis = embedded_scf.mol.basis  # can be anything
        projected_mol.charge = embedded_scf.mol.charge
        projected_mol.spin = embedded_scf.mol.spin
        projected_mol.build()
        projected_mf = scf.UKS(projected_mol)
        n_act_proj_aos = projected_mol.aoslice_by_atom()[self._n_active_atoms - 1][-1]
        logger.debug(f"{n_act_proj_aos=}")

        self.projected_overlap = projected_mf.get_ovlp(embedded_scf.mol)[
            :n_act_proj_aos, :n_act_proj_aos
        ]
        self.overlap_two_basis = gto.intor_cross(
            "int1e_ovlp_sph", embedded_scf.mol, projected_mol
        )[:n_act_proj_aos, :]
        self.n_act_proj_aos = n_act_proj_aos

        spinless = embedded_scf.mo_coeff.ndim == 2

        if spinless:
            localised_virts = self._localize_virtual_spin(
                embedded_scf.mo_occ, embedded_scf.mo_coeff, embedded_scf.get_fock()
            )
            embedded_scf.mo_coeff = localised_virts[0]
            self.shells = localised_virts[1]
            self.singular_values = localised_virts[2]
        else:
            localised_virts_alpha = self._localize_virtual_spin(
                embedded_scf.mo_occ[0],
                embedded_scf.mo_coeff[0],
                embedded_scf.get_fock()[0],
            )
            localised_virts_beta = self._localize_virtual_spin(
                embedded_scf.mo_occ[1],
                embedded_scf.mo_coeff[1],
                embedded_scf.get_fock()[1],
            )
            embedded_scf.mo_coeff = np.array(
                [localised_virts_alpha[0], localised_virts_beta[0]]
            )

            self.shells = (localised_virts_alpha[1], localised_virts_beta[1])
            self.singular_values = (localised_virts_alpha[2], localised_virts_beta[2])

        logger.debug("Completed Concentric Localization.")
        logger.debug(f"{self.shells=}")
        logger.debug(f"{self.singular_values=}")
        return embedded_scf

    def _localize_virtual_spin(
        self, occ: np.ndarray, mo_coeff: np.ndarray, fock_operator: np.ndarray
    ) -> np.ndarray:
        """Run concentric localization for each spin separately.

        NOTE: These cant be done together as the number of occupied orbitals may be different between the two spins.

        Args:
            occ (np.ndarry): MO occupancy
            mo_coeff (np.ndarry): MO coefficient matrix
            fock_operator (np.ndarray): Fock operator for one spin

        Returns:
            np.ndarray: The update MO coefficient matrix
        """
        logger.debug("Running concentric localiztion for single spin.")

        effective_virt = mo_coeff[:, occ == 0]
        logger.debug(f"N effective virtuals: {effective_virt.shape}")

        left = (
            np.linalg.inv(self.projected_overlap)
            @ self.overlap_two_basis
            @ effective_virt
        )
        _, sigma, right_vectors = np.linalg.svd(
            np.swapaxes(left, -1, -2) @ self.overlap_two_basis @ effective_virt
        )
        logger.debug(f"Singular values: {sigma}")

        # record singular values for analysis
        singular_values = []
        singular_values.append(sigma)

        c_total = mo_coeff[:, occ > 0]

        logger.debug(f"Initial {c_total.shape=} (nocc)")
        logger.debug(f"{self.n_act_proj_aos=}")
        shell_size = np.sum(sigma[: self.n_act_proj_aos] >= 1e-15)
        logger.debug(f"{shell_size=}")

        right_vectors = np.swapaxes(right_vectors, -1, -2)
        v_span, v_ker = np.split(
            right_vectors, [shell_size], axis=-1
        )  # 0 but instability

        logger.debug(f"{v_span.shape=}")
        logger.debug(f"{v_ker.shape=}")

        c_ispan = effective_virt @ v_span
        c_iker = effective_virt @ v_ker

        c_total = np.concatenate((c_total, c_ispan), axis=-1)

        # keep track of the number of orbitals in each shell
        shells = []
        shells.append(c_total.shape[-1])
        logger.debug("Created 0th shell.")

        if v_ker.shape[-1] == 0:
            logger.debug("No kernel for 0th shell, cannot perform CL.")
            logger.debug(
                "This is expected for molecules with majority active MOs occupied."
            )
        elif v_ker.shape[-1] == 1:
            logger.debug(
                "Kernel is 1 for 0th shell, ending CL as cannot perform SVD of vector."
            )
            c_total = np.concatenate((c_total, c_iker), axis=-1)
            shells.append(c_total.shape[-1])
        else:
            # why use the overlap for the first shell and then the fock for the rest?
            for ishell in range(0, self.max_shells):
                logger.debug("Beginning Concentric Localization Iteration")
                logger.debug(f"Shell {ishell}.")

                logger.debug(
                    f"{c_total.shape=}, {fock_operator.shape=}, {c_iker.shape=}"
                )
                _, sigma, right_vectors = np.linalg.svd(
                    np.swapaxes(c_total, -1, -2) @ fock_operator @ c_iker
                )
                logger.debug(f"Singular values: {sigma}")
                singular_values.append(sigma)
                logger.debug(f"{right_vectors.shape=}")

                shell_size = np.sum(sigma[: self.n_act_proj_aos] >= 1e-15)
                logger.debug(f"{shell_size=}")
                if shell_size == 0:
                    logger.debug("Empty shell %s, ending CL.", ishell)
                    c_total = np.concatenate((c_total, c_iker), axis=-1)
                    break

                right_vectors = np.swapaxes(right_vectors, -1, -2)
                v_span, v_ker = np.split(
                    right_vectors, [shell_size], axis=-1
                )  # 0 but instability

                logger.debug(f"{v_span.shape=}")
                logger.debug(f"{v_ker.shape=}")

                # span must be done first as both need to use old c_iker
                c_ispan = c_iker @ v_span
                c_total = np.concatenate((c_total, c_ispan), axis=-1)
                shells.append(c_total.shape[-1])

                if v_ker.shape[-1] > 1:
                    logger.debug("Kernel dimension is greater than 1, continuing CL.")
                    # in-place update
                    c_iker = c_iker @ v_ker
                elif v_ker.shape[-1] == 1:
                    c_iker = c_iker @ v_ker
                    logger.debug(
                        "Kernel is 1, ending CL as cannot perform SVD of vector.",
                    )
                    c_total = np.concatenate((c_total, c_iker), axis=-1)
                    shells.append(c_total.shape[-1])
                    break
                else:
                    logger.debug(
                        "Ending Concentric Localization - All virtual MOs localized."
                    )
                    break

                if ishell >= self.max_shells:
                    logger.debug("Max shells reached, not localizing further virtuals.")
                    c_total = np.concatenate((c_total, c_iker), axis=-1)
                    shells.append(c_total.shape[-1])
                    break

        logger.debug(f"Shell indices: {shells}")

        mo_coeff = c_total

        logger.debug(f"{mo_coeff, shells, singular_values}")

        return mo_coeff, shells, singular_values



"""Perform Huzinaga RHF with PySCF."""

import logging
from typing import Optional

import numpy as np
import scipy as sp
from pyscf import dft, scf
from pyscf.lib import StreamObject, diis

logger = logging.getLogger(__name__)


def calculate_hf_energy(
    scf_method, dft_potential, density_matrix, vhf, huzinaga_op_std
) -> float:
    """Calculate the Hartree-Fock Energy.

    Args:
        scf_method (StreamObject): PySCF HF method
        dft_potential (np.ndarray): DFT embedding potential
        density_matrix (np.ndarray): Embedded region density matrix (updates each cycle)
        vhf (np.ndarray): Mean field potential
        huzinaga_op_std (np.ndarray): Huzinaga Fock operator

    Returns:
        float: Hartree-fock energy
    """
    # Find RHF energy
    hamiltonian = scf_method.get_hcore() + dft_potential + 0.5 * vhf + huzinaga_op_std
    return np.einsum("...ij,...ji->...", hamiltonian, density_matrix)


def calculate_ks_energy(
    scf_method, dft_potential, density_matrix, huzinaga_op_std
) -> float:
    """Calculate the Hartree-Fock Energy.

    Args:
        scf_method (StreamObject): PySCF Kohn-sham method
        dft_potential (np.ndarray): DFT embedding potential
        density_matrix (np.ndarray): Embedded region density matrix (updates each cycle)
        huzinaga_op_std (np.ndarray): Huzinaga Fock operator

    Returns:
        float: Kohn-sham energy
    """
    logger.debug("Calculating Huzinaga KS energy")
    logger.debug(f"{dft_potential.shape=}")
    logger.debug(f"{density_matrix.shape=}")
    logger.debug(f"{huzinaga_op_std.shape=}")

    vhf_updated = scf_method.get_veff(dm=density_matrix)
    rks_energy = vhf_updated.ecoul + vhf_updated.exc
    rks_energy += np.einsum(
        "...ij, ...ji->...",
        density_matrix,
        (scf_method.get_hcore() + huzinaga_op_std + dft_potential),
    )
    return rks_energy


def huzinaga_scf(
    scf_method: StreamObject,
    dft_potential: np.ndarray,
    dm_environment: np.ndarray,
    dm_conv_tol: float = 1e-6,
    dm_initial_guess: Optional[np.ndarray] = None,
    use_DIIS: Optional[bool] = True,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, bool]:
    """Manual RHF calculation that is implemented using the huzinaga operator.

    Note this function uses lowdin (symmetric) orthogonalization only! (PySCF sometimes uses meta-lowdin and NAO). Also
    the intial density matrix guess is based on the modified core Hamilotnian (containing projector and DFT potential)
    PySCF has other methods for initial guess that aren't available here. Manual guess can also be given).
    TODO: make a subclass from PySCF RHF object. Can include all this functionality there. Problems in this function
    can occur due to DIIS and other clever PySCF methods not being available.

    Args:
        scf_method (StreamObjecty):PySCF RHF object (containing info about max cycles and convergence tolerence)
        dft_potential (np.ndarray): DFT active and environment two body terms - DFT active environemnt two body term
        dm_environment (np.ndarray): Density matrix of the environment.
        dm_conv_tol (float): density matrix convergence tolerance.
        dm_initial_guess (np.ndarray): Optional initial guess density matrix.
        use_DIIS (bool): whether to use  Direct Inversion in the Iterative Subspace (DIIS) method
    Returns:
        mo_coeff_std (np.ndarray): Optimized C_matrix (columns are optimized moelcular orbtials)
        mo_energy (np.ndarray): 1D array of molecular orbital energies
        density_matrix (np.ndarray): Converged density matrix
        huzinaga_op_std (np.ndarray): Huzinaga operator in standard basis (same basis as Fock operator).
        conv_flag (bool): Flag to indicate whether SCF has converged or not
    """
    logger.debug("Initializising Huzinaga HF calculation")
    s_mat = scf_method.get_ovlp()
    logger.debug(f"{s_mat.shape=}")
    s_neg_half = sp.linalg.fractional_matrix_power(s_mat, -0.5)

    adiis = diis.DIIS() if use_DIIS else None

    # there are many more unrestricted types than restricted
    unrestricted = not isinstance(scf_method, (scf.rhf.RHF, dft.rks.RKS))

    if unrestricted and dm_environment.ndim != 3:
        raise ValueError(
            "Unrestricted calculation requires stacked dm_environment shape (2xMxM)."
        )

    if unrestricted:
        dm_env_S = np.array([dm_environment[0] @ s_mat, dm_environment[1] @ s_mat])
    else:
        dm_env_S = dm_environment @ s_mat

    # Create an initial dm if needed.
    if dm_initial_guess is None:
        fock = scf_method.get_hcore() + dft_potential

        if unrestricted:
            fds_alpha = fock[0] @ dm_env_S[0]
            fds_beta = fock[1] @ dm_env_S[1]
            huzinaga_op_std = np.array(
                [
                    -(fds_alpha + fds_alpha.T),
                    -(fds_beta + fds_beta.T),
                ]
            )
        else:
            fds = fock @ dm_env_S
            # Cant use T as restricted with spin has split DFT potential
            huzinaga_op_std = -0.5 * (fds + np.swapaxes(fds, -1, -2))

        fock += huzinaga_op_std
        # Create the orthogonal fock operator
        fock_ortho = s_neg_half @ fock @ s_neg_half
        mo_energy, mo_coeff_ortho = np.linalg.eigh(fock_ortho)
        mo_coeff_std = s_neg_half @ mo_coeff_ortho
        mo_occ = scf_method.get_occ(mo_energy, mo_coeff_std)
        dm_initial_guess = scf_method.make_rdm1(mo_coeff=mo_coeff_std, mo_occ=mo_occ)

    density_matrix = dm_initial_guess
    conv_flag = False
    scf_energy_prev = 0

    for i in range(scf_method.max_cycle):
        # build fock matrix
        vhf = scf_method.get_veff(dm=density_matrix)
        fock = scf_method.get_hcore() + dft_potential + vhf

        if unrestricted:
            fds_alpha = fock[0] @ dm_env_S[0]
            fds_beta = fock[1] @ dm_env_S[1]
            huzinaga_op_std = np.array(
                [-(fds_alpha + fds_alpha.T), -(fds_beta + fds_beta.T)]
            )
        else:
            fds = fock @ dm_env_S
            huzinaga_op_std = -0.5 * (fds + np.swapaxes(fds, -1, -2))
        fock += huzinaga_op_std

        if use_DIIS and (i > 1):
            # DIIS update of Fock matrix
            fock = adiis.update(fock)

        fock_ortho = s_neg_half @ fock @ s_neg_half

        mo_energy, mo_coeff_ortho = np.linalg.eigh(fock_ortho)
        mo_coeff_std = s_neg_half @ mo_coeff_ortho
        mo_occ = scf_method.get_occ(mo_energy, mo_coeff_std)

        dm_mat_old = density_matrix

        density_matrix = scf_method.make_rdm1(mo_coeff=mo_coeff_std, mo_occ=mo_occ)

        if isinstance(scf_method, (dft.rks.RKS, dft.uks.UKS)):
            # Find RKS energy
            scf_energy = calculate_ks_energy(
                scf_method, dft_potential, density_matrix, huzinaga_op_std
            )
        elif isinstance(scf_method, (scf.rhf.RHF, scf.uhf.UHF)):
            hamiltonian = (
                scf_method.get_hcore() + dft_potential + 0.5 * vhf + huzinaga_op_std
            )
            scf_energy = np.einsum("...ij,...ji->...", hamiltonian, density_matrix)
        else:
            raise TypeError("Cannot run Huzinaga SCF with type %s", type(scf_method))

        # check convergence
        # use max difference so that this works for unrestricted
        run_diff = np.max(np.abs(scf_energy - scf_energy_prev))
        norm_dm_diff = np.max(
            np.linalg.norm(density_matrix - dm_mat_old, axis=(-2, -1))
        )

        if (run_diff < scf_method.conv_tol) and (norm_dm_diff < dm_conv_tol):
            conv_flag = True
            logger.debug("Huzinaga SCF converged in cycle %s", i)
            break

        scf_energy_prev = scf_energy

    if conv_flag is False:
        logger.warning("Huzinaga SCF has NOT converged.")

    return mo_coeff_std, mo_energy, density_matrix, huzinaga_op_std, conv_flag



"""Module containg the NbedDriver Class."""

import logging
from functools import cached_property
from json import dump as jdump
from typing import Optional, Union

import numpy as np
from numpy.typing import NDArray
from pyscf import cc, dft, fci, gto, qmmm, scf
from pyscf.lib import StreamObject

from nbed.localizers import (
    BOYSLocalizer,
    ConcentricLocalizer,
    IBOLocalizer,
    OccupiedLocalizer,
    PMLocalizer,
    SPADELocalizer,
)

from .config import LocalizerEnum, NbedConfig, ProjectorEnum
from .ham_builder import HamiltonianBuilder
from .scf import energy_elec
from .scf.huzinaga_scf import huzinaga_scf

# Create the Logger
logger = logging.getLogger(__name__)


class NbedDriver:
    """Function to return the embedding Qubit Hamiltonian.

    Args:
        config (NbedConfig): A validated config model.

    Attributes:
        _global_fci (StreamObject): A Qubit Hamiltonian of some kind
        e_act (float): Active energy from subsystem DFT calculation
        e_env (float): Environment energy from subsystem DFT calculation
        two_e_cross (float): two electron energy from cross terms (includes exchange correlation
                             and Coulomb contribution) of subsystem DFT calculation
        molecular_ham (InteractionOperator): molecular Hamiltonian for active subsystem (projection using mu shift operator)
        classical_energy (float): environment correction energy to obtain total energy (for mu shift method)
        molecular_ham (InteractionOperator): molecular Hamiltonian for active subsystem (projection using huzianga operator)
    """

    def __init__(self, config: NbedConfig):
        """Initialise NbedDriver."""
        logger.debug("Initialising NbedDriver with config:")
        logger.debug(config.model_dump_json())
        self.config = config
        self.localized_system: OccupiedLocalizer
        self.two_e_cross: np.typing.NDArray
        self.electron: int
        self.mu: dict = None
        self.huzinaga: dict = None

        self._restricted_scf = False
        # if config.force_unrestricted:
        #     logger.debug("Forcing unrestricted SCF")
        #     self._restricted_scf = False
        # elif self.config.charge % 2 == 1 or self.config.spin != 0:
        #     logger.debug("Open shells, using unrestricted SCF.")
        #     self._restricted_scf = False
        # else:
        #     logger.debug("Closed shells, using restricted SCF.")
        #     self._restricted_scf = True

        # if we have values for all three, assume we want to run qmmm
        self.run_qmmm = None not in [
            config.mm_charges,
            config.mm_coords,
            config.mm_radii,
        ]

    def _build_mol(self) -> gto.mole:
        """Function to build PySCF molecule.

        Returns:
            full_mol (gto.mol): built PySCF molecule object
        """
        logger.debug("Constructing molecule.")
        logger.info("Molecule input geometry: %s", self.config.geometry)
        # geometry is raw xyz string
        full_mol = gto.Mole(
            atom=self.config.geometry[3:],
            basis=self.config.basis,
            charge=self.config.charge,
            unit=self.config.unit,
            spin=self.config.spin,
        ).build()
        logger.debug("Molecule built.")
        return full_mol

    @cached_property
    def _global_hf(self, **hf_kwargs) -> StreamObject:
        """Run full system Hartree-Fock."""
        logger.debug("Running full system HF.")
        mol_full = self._build_mol()
        # run Hartree-Fock
        global_hf = scf.UHF(mol_full, **hf_kwargs)
        global_hf.conv_tol = self.config.convergence
        global_hf.max_memory = self.config.max_ram_memory
        global_hf.max_cycle = self.config.max_hf_cycles
        global_hf.kernel()
        logger.info(f"Global HF: {global_hf.e_tot}")

        return global_hf

    @cached_property
    def _global_ccsd(self, **ccsd_kwargs) -> StreamObject:
        """Function to run full molecule CCSD calculation."""
        logger.debug("Running full system CC.")
        # run CCSD after HF

        global_cc = cc.CCSD(self._global_hf, **ccsd_kwargs)
        global_cc.conv_tol = self.config.convergence
        global_cc.max_memory = self.config.max_ram_memory
        global_cc.run()
        logger.info(f"Global CCSD: {global_cc.e_tot}")

        return global_cc

    @cached_property
    def _global_fci(self, **fci_kwargs) -> StreamObject:
        """Function to run full molecule FCI calculation.

        WARNING: FACTORIAL SCALING IN BASIS STATES!
        """
        logger.debug("Running full system FCI.")
        # run FCI after HF
        global_fci = fci.FCI(self._global_hf, **fci_kwargs)
        global_fci.conv_tol = self.config.convergence
        global_fci.max_memory = self.config.max_ram_memory
        global_fci.run()
        logger.info(f"Global FCI: {global_fci.e_tot}")

        return global_fci

    @cached_property
    def _global_ks(self, **ks_kwargs) -> StreamObject:
        """Method to run full cheap molecule USK DFT calculation.

        Note this is necessary to perform localization procedure.
        """
        logger.debug("Running full system KS DFT.")
        mol_full = self._build_mol()
        global_ks = dft.UKS(mol_full, **ks_kwargs)
        logger.debug(f"{type(global_ks)=}")
        global_ks.conv_tol = self.config.convergence
        global_ks.xc = self.config.xc_functional
        global_ks.max_memory = self.config.max_ram_memory
        global_ks.max_cycle = self.config.max_dft_cycles

        if self.run_qmmm:
            logger.debug(
                "QM/MM: running full system KS DFT in presence of point charges."
            )
            global_ks = qmmm.mm_charge(
                global_ks,
                self.config.mm_coords,
                self.config.mm_charges,
                self.config.mm_radii,
            )

        global_ks.kernel()
        logger.debug(f"{global_ks.mo_coeff.shape=}")
        logger.debug(f"{global_ks.mo_occ.shape=}")
        logger.debug(f"{global_ks.get_veff().shape=}")
        logger.debug(f"{global_ks.get_hcore().shape=}")
        logger.info(f"Global USK: {global_ks.e_tot}")

        if global_ks.converged is not True:
            logger.warning("(cheap) global DFT calculation has NOT converged!")

        return global_ks

    def _localize(self) -> OccupiedLocalizer:
        """Run the localizer class."""
        logger.debug(f"Getting localized system using {self.config.localization}.")

        match self.config.localization:
            case LocalizerEnum.SPADE:
                return SPADELocalizer(
                    self._global_ks,
                    self.config.n_active_atoms,
                    max_shells=self.config.max_shells,
                    n_mo_overwrite=self.n_mo_overwrite,
                )
            case LocalizerEnum.BOYS:
                return BOYSLocalizer(
                    self._global_ks,
                    self.config.n_active_atoms,
                    occ_cutoff=self.config.occupied_threshold,
                    virt_cutoff=self.config.virtual_threshold,
                )
            case LocalizerEnum.IBO:
                return IBOLocalizer(
                    self._global_ks,
                    self.config.n_active_atoms,
                    occ_cutoff=self.config.occupied_threshold,
                    virt_cutoff=self.config.virtual_threshold,
                )
            case LocalizerEnum.PM:
                return PMLocalizer(
                    self._global_ks,
                    self.config.n_active_atoms,
                    occ_cutoff=self.config.occupied_threshold,
                    virt_cutoff=self.config.virtual_threshold,
                )

    def _init_local_hf(self) -> Union[scf.uhf.UHF, scf.ROHF]:
        """Function to build embedded HF object for active subsystem.

        Note this function overwrites the total number of electrons to only include active number.

        Returns:
            local_hf (scf.uhf.UHF or scf.ROHF): embedded Hartree-Fock object.
        """
        logger.debug("Constructing localised HF object.")
        embedded_mol: gto.Mole = self._build_mol()

        # overwrite total number of electrons to only include active system
        if self.localized_system.beta_active_MO_inds is None:
            embedded_mol.nelectron = 2 * len(self.localized_system.active_MO_inds)
            embedded_mol.spin = 0
            self._electron = embedded_mol.nelectron
            local_hf: scf.UHF = scf.UHF(embedded_mol)
        else:
            embedded_mol.nelectron = len(self.localized_system.active_MO_inds) + len(
                self.localized_system.beta_active_MO_inds
            )
            embedded_mol.spin = len(self.localized_system.active_MO_inds) - len(
                self.localized_system.beta_active_MO_inds
            )
            self._electron = embedded_mol.nelectron
            local_hf: scf.uhf.UHF = scf.UHF(embedded_mol)
        logger.debug(f"{embedded_mol.nelectron=}")
        logger.debug(f"{embedded_mol.spin=}")

        if self.run_qmmm:
            logger.debug("QM/MM: running local SCF in presence of point charges.")
            local_hf = qmmm.mm_charge(
                local_hf,
                self.config.mm_coords,
                self.config.mm_charges,
                self.config.mm_radii,
            )

        local_hf.max_memory = self.config.max_ram_memory
        local_hf.conv_tol = self.config.convergence
        local_hf.max_cycle = self.config.max_hf_cycles

        return local_hf

    def _init_local_ks(self, xc_functional: str) -> Union[dft.uks.UKS, dft.ROKS]:
        """Function to build embedded Hartree Fock object for active subsystem.

        Note this function overwrites the total number of electrons to only include active number.

        Args:
            xc_functional (str): XC functional to use in embedded calculation.

        Returns:
            local_ks (pyscf.dft.RKS or pyscf.dft.uks.UKS): embedded Kohn-Sham DFT object.
        """
        logger.debug("Initialising localised RKS object.")
        embedded_mol: gto.Mole = self._build_mol()

        if self.localized_system.beta_active_MO_inds is None:
            # overwrite total number of electrons to only include active system
            embedded_mol.nelectron = 2 * len(self.localized_system.active_MO_inds)
            self._electron = embedded_mol.nelectron
            local_ks: dft.UKS = dft.UKS(embedded_mol)
        else:
            embedded_mol.nelectron = len(self.localized_system.active_MO_inds) + len(
                self.localized_system.beta_active_MO_inds
            )
            self._electron = embedded_mol.nelectron
            local_ks: dft.uks.UKS = scf.UKS(embedded_mol)
        logger.debug(f"{embedded_mol.nelectron=}")
        logger.debug(f"{embedded_mol.spin=}")

        local_ks.max_memory = self.config.max_ram_memory
        local_ks.conv_tol = self.config.convergence
        local_ks.xc = xc_functional

        return local_ks

    def _subsystem_dft(
        self, global_ks, localized_system
    ) -> tuple[float, float, np.typing.NDArray]:
        """Function to perform subsystem USK DFT calculation."""
        logger.debug("Calculating active and environment subsystem terms.")

        def _ks_components(
            ks_system: dft.KohnShamDFT,
            subsystem_dm: np.ndarray,
        ) -> tuple[float, np.ndarray, np.ndarray]:
            """Calculate the components of subsystem energy from a USK DFT calculation.

            For a given density matrix this function returns the electronic energy, exchange correlation energy and
            J,K, V_xc matrices.

            Args:
                ks_system (pyscf.dft.KohnShamDFT): PySCF Kohn-Sham object
                subsystem_dm (np.ndarray): density matrix (to calculate all matrices from)


            Returns:
                e_act (float): Active region energy.
                two_e_term (npt.NDArray): Two electron potential term
                j_mat (npt.NDArray): J_matrix defined by input density matrix
            """
            logger.debug("Finding subsystem USK componenets.")
            # It seems that PySCF lumps J and K in the J array
            # need to access the potential for the right subsystem for unrestricted
            logger.debug(f"{subsystem_dm.shape=}")
            two_e_term = ks_system.get_veff(dm=subsystem_dm)
            j_mat = ks_system.get_j(dm=subsystem_dm)
            # k_mat = np.zeros_like(j_mat) not needed for PySCF.

            # v_xc = two_e_term - j_mat

            if subsystem_dm.ndim == 3:
                dm_tot = subsystem_dm[0] + subsystem_dm[1]
            else:
                dm_tot = subsystem_dm
            logger.debug(f"{dm_tot.shape=}")

            # e_act = (
            #     np.einsum("ij,ji->", ks_system.get_hcore(), dm_tot)
            #     + 0.5 * (np.einsum("ij,ji->", j_tot, dm_tot))
            #     + two_e_term.exc
            # )
            e_act = (
                np.einsum("ij,ji->", ks_system.get_hcore(), dm_tot)
                + two_e_term.ecoul
                + two_e_term.exc
            )

            # if check_E_with_pyscf:
            #     energy_elec_pyscf = global_ks.energy_elec(dm=dm_matrix)[0]
            #     if not np.isclose(energy_elec_pyscf, energy_elec):
            #         raise ValueError("Energy calculation incorrect")
            logger.debug("Subsystem USK components found.")
            logger.debug(f"{e_act=}")
            logger.debug(f"{two_e_term=}")
            return e_act, two_e_term, j_mat

        if localized_system.beta_dm_active is None:
            logger.debug("Using spinless density matrix.")
            dm_act = localized_system.dm_active
            dm_env = localized_system.dm_enviro
        else:
            dm_act = np.array(
                [
                    localized_system.dm_active,
                    localized_system.beta_dm_active,
                ]
            )
            dm_env = np.array(
                [
                    localized_system.dm_enviro,
                    localized_system.beta_dm_enviro,
                ]
            )

        (e_act, two_e_act, j_act) = _ks_components(global_ks, dm_act)
        # logger.debug(e_act, alpha_e_xc_act)
        (e_env, two_e_env, j_env) = _ks_components(global_ks, dm_env)
        # logger.debug(alpha_e_env, alpha_e_xc_env, alpha_ecoul_env)

        # Computing cross subsystem terms
        logger.debug("Calculating two electron cross subsystem energy.")
        total_dm = localized_system.dm_active + localized_system.dm_enviro

        if localized_system.beta_dm_active is not None:
            total_dm += (
                localized_system.beta_dm_active + localized_system.beta_dm_enviro
            )

        two_e_term_total = global_ks.get_veff(dm=total_dm)
        logger.debug(f"{total_dm.shape=}")
        logger.debug(f"{two_e_term_total.shape=}")
        e_xc_total = two_e_term_total.exc

        if localized_system.beta_dm_active is None:
            j_cross = 0.5 * (
                np.einsum("ij,ij", localized_system.dm_active, j_env)
                + np.einsum("ij,ij", localized_system.dm_enviro, j_act)
            )
        else:
            j_cross = 0.5 * (
                np.einsum("ij,ij", localized_system.dm_active, j_env[0])
                + np.einsum("ij,ij", localized_system.dm_enviro, j_act[0])
                + np.einsum("ij,ij", localized_system.dm_active, j_env[1])
                + np.einsum("ij,ij", localized_system.dm_enviro, j_act[1])
                + np.einsum("ij,ij", localized_system.beta_dm_active, j_env[1])
                + np.einsum("ij,ij", localized_system.beta_dm_enviro, j_act[1])
                + np.einsum("ij,ij", localized_system.beta_dm_active, j_env[0])
                + np.einsum("ij,ij", localized_system.beta_dm_enviro, j_act[0])
            )
        logger.debug(f"{j_cross=}")

        # Because of projection we expect kinetic term to be zero
        k_cross = 0.0

        xc_cross = e_xc_total - two_e_act.exc - two_e_env.exc
        logger.debug(f"{e_xc_total=}")
        logger.debug(f"{two_e_act.exc=}")
        logger.debug(f"{two_e_env.exc=}")

        # overall two_electron cross energy
        two_e_cross = j_cross + k_cross + xc_cross

        logger.debug("USK components")
        logger.debug(f"e_act: {e_act}")
        logger.debug(f"e_env: {e_env}")
        logger.debug(f"two_e_cross: {two_e_cross}")
        logger.debug(f"e_nuc: {global_ks.energy_nuc()}")
        return e_act, e_env, two_e_cross

    @cached_property
    def _env_projector(self) -> np.ndarray:
        """Return a projector onto the environment in orthogonal basis."""
        logger.debug("Getting Environment Projector.")
        s_mat = self._global_ks.get_ovlp()
        logger.debug(f"{s_mat.shape=}")
        env_projector_alpha = s_mat @ self.localized_system.dm_enviro @ s_mat

        if self.localized_system.beta_dm_enviro is None:
            env_projector = env_projector_alpha

        else:
            env_projector_beta = s_mat @ self.localized_system.beta_dm_enviro @ s_mat
            env_projector = np.array([env_projector_alpha, env_projector_beta])
        logger.debug(f"{env_projector.shape=}")
        return env_projector

    def _run_emb_ccsd(
        self,
        emb_pyscf_scf_rhf: Union[scf.ROHF, scf.UHF],
        frozen: Optional[list] = None,
    ) -> tuple[cc.CCSD, float]:
        """Function run CCSD on embedded restricted Hartree Fock object.

        Note emb_pyscf_scf_rhf is ROHF object for the active embedded subsystem (defined in localized basis)
        (see get_embedded_rhf method)

        Args:
            emb_pyscf_scf_rhf (scf.ROHF): PySCF restricted Hartree Fock object of active embedded subsystem
            frozen (List): A path to an .xyz file describing molecular geometry.

        Returns:
            ccsd (cc.CCSD): PySCF CCSD object
            e_ccsd_corr (float): electron correlation CCSD energy
        """
        return run_emb_ccsd(
            emb_pyscf_scf_rhf,
            frozen,
            self.config.convergence,
            self.config.max_ram_memory,
        )

    def _run_emb_fci(
        self,
        emb_pyscf_scf_rhf: Union[scf.ROHF, scf.UHF],
        frozen: Optional[list] = None,
    ) -> fci.FCI:
        """Function run FCI on embedded restricted Hartree Fock object.

        Note emb_pyscf_scf_rhf is ROHF object for the active embedded subsystem (defined in localized basis)
        (see get_embedded_rhf method)

        Args:
            emb_pyscf_scf_rhf (scf.ROHF): PySCF restricted Hartree Fock object of active embedded subsystem
            frozen (List): A path to an .xyz file describing moleclar geometry.

        Returns:
            fci_scf (fci.FCI): PySCF FCI object
        """
        return run_emb_fci(
            emb_pyscf_scf_rhf,
            frozen,
            self.config.convergence,
            self.config.max_ram_memory,
        )

    def _mu_embed(
        self, localized_scf: StreamObject, embedding_potential: np.ndarray
    ) -> tuple[StreamObject, np.ndarray]:
        """Embed using the Mu-shift projector.

        Args:
            localized_scf (StreamObject): A PySCF scf method with the correct number of electrons for the active region.
            embedding_potential (np.ndarray): Potential calculated from two electron terms in dft.

        Returns:
            np.ndarray: Matrix form of the embedding potential.
            StreamObject: The embedded scf object.
        """
        logger.debug("Running mu embedded scf calculation.")

        # Modify the energy_elec function to handle different h_cores
        # which we need for different embedding potentials

        v_emb = (self.config.mu_level_shift * self._env_projector) + embedding_potential
        logger.debug(f"{v_emb.shape=}")

        if v_emb.ndim == 3:
            localized_scf.energy_elec = lambda *args: energy_elec(localized_scf, *args)

        logger.debug(f"{v_emb.shape=}")
        logger.debug(f"{self._env_projector.shape=}")
        logger.debug(f"{embedding_potential.shape=}")
        hcore_std = localized_scf.get_hcore
        logger.debug(f"{hcore_std().shape=}")
        localized_scf.get_hcore = lambda *args: hcore_std(*args) + v_emb
        # veff_std = localized_scf.get_veff
        # localized_scf.get_veff = lambda *args: veff_std(*args) + v_emb
        logger.debug(f"embedded hcore shape {localized_scf.get_hcore().shape}")

        localized_scf.kernel()
        logger.info(
            f"Embedded scf energy MU_SHIFT: {localized_scf.e_tot}, converged: {localized_scf.converged}"
        )

        return localized_scf, v_emb

    def _huzinaga_embed(
        self,
        active_scf: StreamObject,
        embedding_potential: np.ndarray,
        dmat_initial_guess: Optional[tuple[np.ndarray]] = None,
    ) -> tuple[StreamObject, np.ndarray]:
        """Embed using Huzinaga projector.

        Args:
            active_scf (StreamObject): A PySCF scf method with the correct number of electrons for the active region.
            embedding_potential (np.ndarray): Potential calculated from two electron terms in dft.
            dmat_initial_guess (bool): If True, use the initial guess for the density matrix.

        Returns:
            np.ndarray: Matrix form of the embedding potential.
            StreamObject: The embedded scf object.
        """
        logger.debug("Starting Huzinaga embedding method.")
        # We need to run our own SCF method here to update the potential.
        if self.localized_system.beta_dm_enviro is None:
            total_enviro_dm = self.localized_system.dm_enviro
        else:
            total_enviro_dm = np.array(
                [
                    self.localized_system.dm_enviro,
                    self.localized_system.beta_dm_enviro,
                ]
            )
        localized_scf = active_scf
        (
            c_active_embedded,
            mo_embedded_energy,
            dm_active_embedded,
            huzinaga_op_std,
            huz_scf_conv_flag,
        ) = huzinaga_scf(
            localized_scf,
            embedding_potential,
            total_enviro_dm,
            dm_conv_tol=1e-6,
            dm_initial_guess=dmat_initial_guess,
        )

        logger.debug(f"{c_active_embedded=}")

        # write results to pyscf object
        logger.debug("Writing results to PySCF object.")
        hcore_std = localized_scf.get_hcore()
        v_emb = huzinaga_op_std + embedding_potential
        localized_scf.get_hcore = lambda *args: hcore_std + v_emb

        if self.localized_system.spinless is False:
            localized_scf.energy_elec = lambda *args: energy_elec(localized_scf, *args)

        localized_scf.mo_coeff = c_active_embedded
        localized_scf.mo_occ = localized_scf.get_occ(
            mo_embedded_energy, c_active_embedded
        )
        logger.debug(f"{localized_scf.mo_occ=}")
        localized_scf.mo_energy = mo_embedded_energy
        localized_scf.e_tot = localized_scf.energy_tot(dm=dm_active_embedded)
        # localized_scf.conv_check = huz_scf_conv_flag
        localized_scf.converged = huz_scf_conv_flag

        logger.info(f"Embedded scf energy HUZINAGA: {localized_scf.e_tot}")
        return localized_scf, v_emb

    def _delete_spin_environment(
        self,
        projector: ProjectorEnum,
        n_env_mo: int,
        mo_coeff: np.ndarray,
        mo_energy: np.ndarray,
        mo_occ: np.ndarray,
        environment_projector: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Remove enironment orbit from embedded ROHF object.

        This function removes (in fact deletes completely) the molecular orbitals
        defined by the environment of the localized system

        Args:
            projector (ProjectorEnum): The projector used to embed the system.
            n_env_mo (int): The number of molecular orbitals in the environment.
            mo_coeff (np.ndarray): The molecular orbitals.
            mo_energy (np.ndarray): The molecular orbital energies.
            mo_occ (np.ndarray): The molecular orbital occupation numbers.
            environment_projector (np.ndarray): Matrix to project mo_coeff onto environment.

        Returns:
            embedded_rhf (StreamObject): Returns input, but with environment orbitals deleted
        """
        logger.debug("Deleting environment for spin.")
        logger.debug(f"{projector=}")
        logger.debug(f"{n_env_mo=}")
        logger.debug(f"{mo_coeff.shape=}")
        logger.debug(f"{mo_energy=}")
        logger.debug(f"{mo_occ=}")
        logger.debug(f"{environment_projector.shape=}")

        match projector:
            case ProjectorEnum.HUZ:
                # MOs which have the greatest overlap with the
                overlap = np.einsum(
                    "ij, ki -> i",
                    mo_coeff.swapaxes(-1, -2),
                    environment_projector @ mo_coeff,
                )
                overlap_by_size = overlap.argsort()[::-1]
                logger.debug(f"{overlap_by_size=}")
                frozen_enviro_orb_inds = overlap_by_size[:n_env_mo]

            case ProjectorEnum.MU:
                # Orbitals which have been shifted to have energy mu are removed
                shift = mo_coeff.shape[-1] - n_env_mo
                frozen_enviro_orb_inds = [
                    mo_i for mo_i in range(shift, mo_coeff.shape[-1])
                ]

        active_MOs_occ_and_virt_embedded = [
            mo_i
            for mo_i in range(mo_coeff.shape[-1])
            if mo_i not in frozen_enviro_orb_inds
        ]

        logger.info(
            f"Orbital indices for embedded system: {active_MOs_occ_and_virt_embedded}"
        )
        logger.info(
            f"Orbital indices removed from embedded system: {frozen_enviro_orb_inds}"
        )

        # delete enviroment orbitals and associated energies
        # overwrites varibles keeping only active part (both occupied and virtual)
        active_mo_coeff = mo_coeff[:, active_MOs_occ_and_virt_embedded]
        active_mo_energy = mo_energy[active_MOs_occ_and_virt_embedded]
        active_mo_occ = mo_occ[: len(active_MOs_occ_and_virt_embedded)]

        logger.debug("Spin environment deleted.")
        logger.debug(f"{active_mo_coeff=}")
        logger.debug(f"{active_mo_energy=}")
        logger.debug(f"{active_mo_occ=}")
        return active_mo_coeff, active_mo_energy, active_mo_occ

    def _delete_environment(
        self, projector: ProjectorEnum, scf: StreamObject
    ) -> StreamObject:
        """Remove enironment orbit from embedded ROHF object.

        This function removes (in fact deletes completely) the molecular orbitals
        defined by the environment of the localized system

        Args:
            projector (ProjectorEnum): The projector used to embed the system.
            scf (StreamObject): The embedded SCF object.

        Returns:
            StreamObject: Returns input, but with environment orbitals deleted.
        """
        logger.debug("Deleting environment from SCF object.")

        if self.localized_system.beta_enviro_MO_inds is None:
            n_env_mos = len(self.localized_system.enviro_MO_inds)
            scf.mo_coeff, scf.mo_energy, scf.mo_occ = self._delete_spin_environment(
                projector,
                n_env_mos,
                scf.mo_coeff,
                scf.mo_energy,
                scf.mo_occ,
                self._env_projector,
            )
        else:
            alpha_n_env_mos = len(self.localized_system.enviro_MO_inds)
            beta_n_env_mos = len(self.localized_system.beta_enviro_MO_inds)
            mo_coeff = np.array([None, None])
            mo_energy = np.array([None, None])
            mo_occ = np.array([None, None])
            (
                mo_coeff[0],
                mo_energy[0],
                mo_occ[0],
            ) = self._delete_spin_environment(
                projector,
                alpha_n_env_mos,
                scf.mo_coeff[0],
                scf.mo_energy[0],
                scf.mo_occ[0],
                self._env_projector[0],
            )
            (mo_coeff[1], mo_energy[1], mo_occ[1]) = self._delete_spin_environment(
                projector,
                beta_n_env_mos,
                scf.mo_coeff[1],
                scf.mo_energy[1],
                scf.mo_occ[1],
                self._env_projector[1],
            )

            # Need to do it this way or there are broadcasting issues
            scf.mo_coeff = np.array(
                [mo_coeff[0], mo_coeff[1]]
            )  # np.array([mo_coeff[0], mo_coeff[1]])
            scf.mo_energy = np.array(
                [mo_energy[0], mo_energy[1]]
            )  # np.array([mo_energy[0], mo_energy[1]])
            scf.mo_occ = np.array(
                [mo_occ[0], mo_occ[1]]
            )  # np.array([mo_occ[0], mo_occ[1]])

        logger.debug("Environment deleted.")
        return scf

    def _dft_in_dft(self, projection_method: ProjectorEnum) -> dict:
        """Return energy of DFT in DFT embedding.

        Note run_mu_shift (bool) and run_huzinaga (bool) flags define which method to use (can be both)
        This is done when object is initialized.

        Args:
            driver (NbedDriver): A driver object.
            projection_method (callable): Embedding method to use (mu or huzinaga).

        Returns:
            dict: DFT-in-DFT embedding results.
        """
        return dft_in_dft(self, projection_method)

    def embed(
        self,
        init_huzinaga_rhf_with_mu: bool = False,
        n_mo_overwrite: tuple[int | None, int | None] = (None, None),
    ) -> None:
        """Run embedded scf calculation.

        Args:
            init_huzinaga_rhf_with_mu (bool): Will run mu-shift projector even when input projector='huzinaga'.
            n_mo_overwrite (tuple[int, int]): Enforces a specific number of MOs are included in the active region. Used for ACE-of-SPADE reaction path localization.
        """
        logger.debug("Embedding molecule.")
        self.e_nuc = self._global_ks.energy_nuc()

        if n_mo_overwrite is not None and n_mo_overwrite != (None, None):
            logger.debug(
                "Setting n_mo_overwrite with value from embed args %s", n_mo_overwrite
            )
            self.n_mo_overwrite = n_mo_overwrite
        else:
            logger.debug("Setting n_mo_overwrite with value from config.")
            self.n_mo_overwrite = self.config.n_mo_overwrite

        self.localized_system = self._localize()
        logger.info("Indices of embedded electrons:")
        logger.info(self.localized_system.active_MO_inds)
        logger.info(self.localized_system.enviro_MO_inds)

        # Run subsystem DFT (calls localized rks)
        self.e_act, self.e_env, self.two_e_cross = self._subsystem_dft(
            self._global_ks, self.localized_system
        )
        logger.debug("Getting global DFT potential to optimize embedded calc in.")

        total_dm = self.localized_system.dm_active + self.localized_system.dm_enviro
        if self.localized_system.beta_dm_active is not None:
            logger.debug("Adding beta spin density matrix")
            total_dm = np.array(
                [
                    total_dm,
                    self.localized_system.beta_dm_active
                    + self.localized_system.beta_dm_enviro,
                ]
            )

        logger.debug(f"{self._global_ks.get_veff().shape=}")
        logger.debug(f"{self._global_ks.get_veff(dm=total_dm).shape=}")
        g_act_and_env = self._global_ks.get_veff(dm=total_dm)
        logger.debug(f"{total_dm.shape=}")
        logger.debug(f"{g_act_and_env.shape=}")

        if self.localized_system.beta_dm_active is None:
            g_act = self._global_ks.get_veff(dm=self.localized_system.dm_active)
        else:
            g_act = self._global_ks.get_veff(
                dm=np.array(
                    [
                        self.localized_system.dm_active,
                        self.localized_system.beta_dm_active,
                    ]
                )
            )
        embedding_potential = g_act_and_env - g_act
        self.embedding_potential = embedding_potential

        logger.info(f"DFT potential average {np.mean(embedding_potential)}.")

        logger.debug("Beginning Projection.")
        if (
            self.config.projector in [ProjectorEnum.MU, ProjectorEnum.BOTH]
            or init_huzinaga_rhf_with_mu
        ):
            local_hf = self._init_local_hf()
            embedded_scf, v_emb = self._mu_embed(local_hf, embedding_potential)
            self.mu = self.collect_results(embedded_scf, v_emb, ProjectorEnum.MU)

        if self.config.projector in [ProjectorEnum.HUZ, ProjectorEnum.BOTH]:
            local_hf = self._init_local_hf()
            dmat_initial_guess: Optional[tuple[NDArray]] = (
                self.mu["scf"].make_rdm1() if init_huzinaga_rhf_with_mu else None
            )
            embedded_scf, v_emb = self._huzinaga_embed(
                local_hf, embedding_potential, dmat_initial_guess
            )
            self.huzinaga = self.collect_results(embedded_scf, v_emb, ProjectorEnum.HUZ)

        match self.config.projector:
            case ProjectorEnum.MU:
                self.embedded_scf = self.mu["scf"]
                self.classical_energy = self.mu["classical_energy"]
            case ProjectorEnum.HUZ:
                self.embedded_scf = self.huzinaga["scf"]
                self.classical_energy = self.huzinaga["classical_energy"]
            case ProjectorEnum.BOTH:
                logger.warning(
                    "Outputting both mu and huzinaga embedding results as tuple."
                )
                self.embedded_scf = (
                    self.mu["scf"],
                    self.huzinaga["scf"],
                )
                self.classical_energy = (
                    self.mu["classical_energy"],
                    self.huzinaga["classical_energy"],
                )
            case _:
                logger.debug("Projector did not match any know case.")
                logger.warning("Not assigning embedded_scf or classial_energy")

        if filename := self.config.savefile is not None:
            logger.debug("Saving results to file %s", filename)
            with open(filename, "w") as f:
                jdump({"mu": self.mu, "huzinaga": self.huzinaga}, f)

        logger.info("Embedding complete.")

    def collect_results(
        self, embedded_scf: StreamObject, v_emb: NDArray, projector: ProjectorEnum
    ) -> dict:
        """Build a results dictionary.

        Args:
            embedded_scf (StreamObject): An embedded pyscf scf object.
            v_emb (NDArray): Embedding Potential
            projector (ProjectorEnum): Which projector the result should use.

        Returns:
            dict: A dict of results.
        """
        result = {}
        result["scf"] = embedded_scf.copy()
        result["v_emb"] = v_emb
        result["mo_energies_emb_pre_del"] = result["scf"].mo_energy
        result["scf"] = self._delete_environment(projector, result["scf"])
        result["mo_energies_emb_post_del"] = result["scf"].mo_energy

        logger.info(f"V emb mean {projector}: {np.mean(result['v_emb'])}")

        # calculate correction
        if self.localized_system.beta_dm_active is None:
            result["correction"] = np.einsum(
                "ij,ij", result["v_emb"], self.localized_system.dm_active
            )
            result["beta_correction"] = 0
        else:
            result["correction"] = np.einsum(
                "ij,ij", result["v_emb"][0], self.localized_system.dm_active
            )
            result["beta_correction"] = np.einsum(
                "ij,ij", result["v_emb"][1], self.localized_system.beta_dm_active
            )

        # Virtual localization
        # TODO correlation energy correction???
        if self.config.run_virtual_localization is True:
            logger.debug("Performing virtual localization.")
            result["scf"] = ConcentricLocalizer(
                result["scf"],
                self.config.n_active_atoms,
                max_shells=self.config.max_shells,
            ).localize_virtual()

        result["e_rhf"] = (
            result["scf"].e_tot
            + self.e_env
            + self.two_e_cross
            - result["correction"]
            - result["beta_correction"]
        )
        logger.info(f"ROHF energy: {result['e_rhf']}")

        # classical energy
        result["classical_energy"] = (
            self.e_env
            + self.two_e_cross
            + self.e_nuc
            - result["correction"]
            - result["beta_correction"]
        )
        logger.debug(f"Classical energy: {result['classical_energy']}")

        # Calculate ccsd or fci energy
        if self.config.run_ccsd_emb is True:
            logger.debug("Performing CCSD-in-DFT embedding.")
            ccsd_emb, e_ccsd_corr = self._run_emb_ccsd(result["scf"])
            result["e_ccsd"] = (
                ccsd_emb.e_tot
                + self.e_env
                + self.two_e_cross
                - result["correction"]
                - result["beta_correction"]
            )
            result["ccsd_emb"] = ccsd_emb.e_tot - self.e_nuc

            logger.info(f"CCSD Energy {projector}:\t{result['e_ccsd']}")

        if self.config.run_fci_emb is True:
            logger.debug("Performing FCI-in-DFT embedding.")
            fci_emb = self._run_emb_fci(result["scf"])
            result["e_fci"] = (
                (fci_emb.e_tot)
                + self.e_env
                + self.two_e_cross
                - result["correction"]
                - result["beta_correction"]
            )
            logger.info(f"FCI Energy {projector}:\t{result['e_fci']}")

            result["fci_emb"] = fci_emb.e_tot - self.e_nuc
        result["hf_emb"] = result["scf"].e_tot - self.e_nuc

        if self.config.run_dft_in_dft is True:
            did = self._dft_in_dft(projector)
            result.update(did)

        # Build second quantised Hamiltonian
        hb = HamiltonianBuilder(result["scf"], result["classical_energy"])
        result["second_quantised"] = hb.build()

        logger.debug(f"Found result for {projector}")
        logger.debug(result)

        return result


def run_emb_fci(
    emb_pyscf_scf_rhf: Union[scf.ROHF, scf.UHF],
    frozen: Optional[list] = None,
    convergence: Optional[float] = 1e-6,
    max_ram_memory: Optional[int] = 4000,
) -> fci.FCI:
    """Function run FCI on embedded restricted Hartree Fock object.

    Note emb_pyscf_scf_rhf is ROHF object for the active embedded subsystem (defined in localized basis)
    (see get_embedded_rhf method)

    Args:
        emb_pyscf_scf_rhf (scf.ROHF): PySCF restricted Hartree Fock object of active embedded subsystem
        frozen (List): A path to an .xyz file describing moleclar geometry.
        convergence (float): convergence tolerance.
        max_ram_memory (int): Maximum memory allocation for FCI.

    Returns:
        fci_scf (fci.FCI): PySCF FCI object
    """
    logger.debug("Starting embedded FCI calculation.")
    logger.debug(f"{type(emb_pyscf_scf_rhf)=}")
    logger.debug(f"{frozen=}")
    logger.debug(f"{convergence=}")
    logger.debug(f"{max_ram_memory=}")

    if frozen is None:
        fci_scf = fci.FCI(emb_pyscf_scf_rhf)
    else:
        from pyscf import mcscf

        fci_scf = mcscf.CASSCF(
            emb_pyscf_scf_rhf,
            emb_pyscf_scf_rhf.mol.nelec,
            emb_pyscf_scf_rhf.mol.nao - len(frozen),
        )
        fci_scf.sort_mo(
            [i + 1 for i in range(emb_pyscf_scf_rhf.mol.nao) if i not in frozen]
        )
    fci_scf.conv_tol = convergence
    fci_scf.max_memory = max_ram_memory

    # For UHF, PySCF assumes that hcore is spinless and 2D
    # Because we update hcore for embedding, we need to calculate our own h1e term.
    from functools import reduce

    if np.ndim(hcore := emb_pyscf_scf_rhf.get_hcore()) == 3:
        mo = emb_pyscf_scf_rhf.mo_coeff
        h1e = [
            reduce(np.dot, (mo[0].T, hcore[0], mo[0])),
            reduce(np.dot, (mo[1].T, hcore[1], mo[1])),
        ]
        fci_scf.kernel(h1e=h1e)
    else:
        # kernel function default value is passed in
        fci_scf.kernel()
    logger.info(f"FCI embedding energy: {fci_scf.e_tot}")
    return fci_scf


def run_emb_ccsd(
    emb_pyscf_scf_rhf: Union[scf.ROHF, scf.UHF],
    frozen: Optional[list] = None,
    convergence: float = 1e-6,
    max_ram_memory: int = 4000,
) -> tuple[cc.CCSD, float]:
    """Function run CCSD on embedded restricted Hartree Fock object.

    Note emb_pyscf_scf_rhf is ROHF object for the active embedded subsystem (defined in localized basis)
    (see get_embedded_rhf method)

    Args:
        emb_pyscf_scf_rhf (scf.ROHF): PySCF restricted Hartree Fock object of active embedded subsystem
        frozen (List): A path to an .xyz file describing molecular geometry.
        convergence (float): Convergence threshold.
        max_ram_memory (int): Maximum ram to use in solving.

    Returns:
        ccsd (cc.CCSD): PySCF CCSD object
        e_ccsd_corr (float): electron correlation CCSD energy
    """
    logger.debug("Starting embedded CCSD calculation.")
    ccsd = cc.CCSD(emb_pyscf_scf_rhf, frozen=frozen)
    ccsd.conv_tol = convergence
    ccsd.max_memory = max_ram_memory

    e_ccsd_corr, _, _ = ccsd.kernel()
    logger.info(f"Embedded CCSD energy: {e_ccsd_corr}")
    return ccsd, e_ccsd_corr


def dft_in_dft(driver: "NbedDriver", projection_method: ProjectorEnum) -> dict:
    """Return energy of DFT in DFT embedding.

    Note run_mu_shift (bool) and run_huzinaga (bool) flags define which method to use (can be both)
    This is done when object is initialized.

    Args:
        driver (NbedDriver): A driver object.
        projection_method (callable): Embedding method to use (mu or huzinaga).

    Returns:
        dict: DFT-in-DFT embedding results.
    """
    result = {}
    e_nuc = driver._global_ks.energy_nuc()

    local_rks_same_functional = driver._init_local_ks(driver._global_ks.xc)
    hcore_std = local_rks_same_functional.get_hcore()
    match projection_method:
        case ProjectorEnum.MU:
            result["scf_dft"], result["v_emb_dft"] = driver._mu_embed(
                local_rks_same_functional, driver.embedding_potential
            )
        case ProjectorEnum.HUZ:
            result["scf_dft"], result["v_emb_dft"] = driver._huzinaga_embed(
                local_rks_same_functional, driver.embedding_potential
            )
    result["scf_dft"] = driver._delete_environment(projection_method, result["scf_dft"])

    if driver.localized_system.beta_dm_active is not None:
        y_emb_alpha, y_emb_beta = result["scf_dft"].make_rdm1()

        # calculate correction
        result["dft_correction"] = np.einsum(
            "ij,ij",
            result["v_emb_dft"][0],
            (y_emb_alpha - driver.localized_system.dm_active),
        )

        result["dft_correction_beta"] = np.einsum(
            "ij,ij",
            result["v_emb_dft"][1],
            (y_emb_beta - driver.localized_system.beta_dm_active),
        )

        veff = result["scf_dft"].get_veff(dm=[y_emb_alpha, y_emb_beta])

        rks_e_elec = (
            veff.exc
            + veff.ecoul
            + np.einsum(
                "ij,ij",
                hcore_std,
                y_emb_alpha,
            )
            + np.einsum(
                "ij,ij",
                hcore_std,
                y_emb_beta,
            )
        )

    else:
        y_emb = result["scf_dft"].make_rdm1()

        # calculate correction
        result["dft_correction"] = np.einsum(
            "ij,ij",
            result["v_emb_dft"],
            (y_emb - driver.localized_system.dm_active),
        )
        veff = result["scf_dft"].get_veff(dm=y_emb)
        result["dft_correction_beta"] = 0
        rks_e_elec = veff.exc + veff.ecoul + np.einsum("ij,ij", hcore_std, y_emb)

    result["e_dft_in_dft"] = (
        rks_e_elec
        + driver.e_env
        + driver.two_e_cross
        + result["dft_correction"]
        + result["dft_correction_beta"]
        + e_nuc
    )
    result["emb_dft"] = rks_e_elec

    return result



"""Test SCF Functions."""

from nbed.scf import huzinaga_scf
import pytest
from numpy.typing import NDArray
import numpy as np


@pytest.fixture()
def dft_potential(spinless_driver) -> NDArray:
    return spinless_driver.embedding_potential


@pytest.fixture()
def dm_environment(spinless_driver) -> NDArray:
    return np.array(
        [
            spinless_driver.localized_system.dm_enviro,
            spinless_driver.localized_system.beta_dm_enviro,
        ]
    )


def test_rks_output(water_rks, dft_potential, dm_environment):
    scf_result = huzinaga_scf(
        water_rks, dft_potential=dft_potential[0], dm_environment=dm_environment[0]
    )
    assert np.allclose(scf_result[0].shape, (7, 7))
    assert np.allclose(
        scf_result[1],
        [
            -17.44629099,
            -0.27614116,
            0.37893061,
            0.89022282,
            1.12092664,
            3.32762378,
            3.86532114,
        ],
    )
    assert np.allclose(scf_result[2].shape, (7, 7))
    assert np.allclose(np.mean(scf_result[2]), 0.1822057642580939)
    assert np.allclose(scf_result[3].shape, (7, 7))
    assert np.allclose(np.mean(scf_result[3]), -0.011214890666261626)
    assert np.allclose(scf_result[4], True)


def test_uks_output(water_uks, dft_potential, dm_environment):
    scf_result = huzinaga_scf(
        water_uks, dft_potential=dft_potential, dm_environment=dm_environment
    )
    assert np.allclose(scf_result[0].shape, (2, 7, 7))
    assert np.allclose(
        scf_result[1],
        [
            [
                -17.29060406,
                -0.28451256,
                0.31504139,
                0.60348835,
                1.0520797,
                2.22020625,
                3.8346852,
            ],
            [
                -17.29048252,
                -0.2845074,
                0.31505414,
                0.60339465,
                1.05207751,
                2.22026345,
                3.83468313,
            ],
        ],
    )
    assert np.allclose(scf_result[2].shape, (2, 7, 7))
    assert np.allclose(np.mean(scf_result[2]), 0.09276688041715254)
    assert np.allclose(scf_result[3].shape, (2, 7, 7))
    assert np.allclose(np.mean(scf_result[3]), -0.02251188710459783)
    assert np.allclose(scf_result[4], True)


def test_rhf_output(water_rhf, dft_potential, dm_environment):
    scf_result = huzinaga_scf(
        water_rhf, dft_potential=dft_potential[0], dm_environment=dm_environment[0]
    )
    assert np.allclose(scf_result[0].shape, (7, 7))
    assert np.allclose(
        scf_result[1],
        [
            -19.346243,
            -0.59741322,
            0.12747464,
            0.6132579,
            0.79561917,
            3.56833278,
            4.1655741,
        ],
    )
    assert np.allclose(scf_result[2].shape, (7, 7))
    assert np.allclose(np.mean(scf_result[2]), 0.17985591319811933)
    assert np.allclose(scf_result[3].shape, (7, 7))
    assert np.allclose(np.mean(scf_result[3]), -0.01224642921175508)
    assert np.allclose(scf_result[4], True)


def test_uhf_output(water_uhf, dft_potential, dm_environment):
    scf_result = huzinaga_scf(
        water_uhf, dft_potential=dft_potential, dm_environment=dm_environment
    )
    assert np.allclose(scf_result[0].shape, (2, 7, 7))
    assert np.allclose(
        scf_result[1],
        [
            [
                -19.18005207,
                -0.618383,
                0.07366692,
                0.39496279,
                0.72192366,
                2.44806433,
                4.12874389,
            ],
            [
                -19.17991953,
                -0.6183819,
                0.07366408,
                0.39491023,
                0.72191934,
                2.44812268,
                4.12874047,
            ],
        ],
    )
    assert np.allclose(scf_result[2].shape, (2, 7, 7))
    assert np.allclose(np.mean(scf_result[2]), 0.0920247346776863)
    assert np.allclose(scf_result[3].shape, (2, 7, 7))
    assert np.allclose(np.mean(scf_result[3]), -0.024315876434944768)
    assert np.allclose(scf_result[4], True)



"""File to contain tests of the driver.py script."""

import logging

import numpy as np
import pytest
from numpy import isclose
from pyscf.lib.misc import StreamObject

from nbed.driver import NbedDriver
from nbed.config import NbedConfig, ProjectorEnum
from pydantic import ValidationError

logger = logging.getLogger(__name__)


@pytest.fixture
def mu_driver(nbed_config) -> NbedDriver:
    nbed_config.projector = ProjectorEnum.MU
    driver = NbedDriver(nbed_config)
    driver.embed()
    return driver


@pytest.fixture
def huz_driver(nbed_config) -> NbedDriver:
    nbed_config.projector = ProjectorEnum.HUZ
    driver = NbedDriver(nbed_config)
    driver.embed()
    return driver


@pytest.fixture
def both_driver(nbed_config) -> NbedDriver:
    nbed_config.projector = ProjectorEnum.BOTH
    driver = NbedDriver(nbed_config)
    driver.embed()
    return driver


@pytest.mark.parametrize("driver", ["both_driver", "mu_driver", "huz_driver"])
def test_global_ks(request, driver):
    driver = request.getfixturevalue(driver)
    result = driver._global_ks()
    assert np.isclose(result.e_tot, np.float64(-75.3091447400438))
    assert np.allclose(
        result.energy_elec(),
        (np.float64(-84.59485896172163), np.float64(37.93302591280513)),
    )


@pytest.mark.parametrize("driver", ["both_driver", "mu_driver", "huz_driver"])
def test_global_hf(request, driver):
    driver = request.getfixturevalue(driver)
    result = driver._global_hf()
    assert np.isclose(result.energy_nuc(), np.float64(9.285714221677825))
    assert np.isclose(result.e_tot, -74.96099960129165)
    assert np.allclose(
        result.energy_elec(),
        (np.float64(-84.24671382296947), np.float64(38.288174841671974)),
    )


@pytest.mark.parametrize("driver", ["both_driver", "mu_driver", "huz_driver"])
def test_global_ccsd(request, driver):
    driver = request.getfixturevalue(driver)
    result = driver._global_ccsd()
    assert np.isclose(result.e_tot, -75.0090124134578)
    assert np.isclose(result.e_corr, -0.04801281045273269)


@pytest.mark.parametrize("driver", ["both_driver", "mu_driver", "huz_driver"])
def test_global_fci(request, driver):
    driver = request.getfixturevalue(driver)
    result = driver._global_fci()
    assert np.isclose(result.e_tot, np.float64(-75.00912605315143))
    from nbed.driver import run_emb_fci

    emb_result = run_emb_fci(driver._global_hf)
    assert np.isclose(emb_result.e_tot, np.float64(-75.00912605315143))


def test_restricted_dft_in_dft(mu_driver, huz_driver):
    mu_did = mu_driver._dft_in_dft(ProjectorEnum.MU)
    huz_did = huz_driver._dft_in_dft(ProjectorEnum.HUZ)
    assert np.isclose(mu_did["e_dft_in_dft"], mu_driver._global_ks().e_tot)
    assert np.isclose(huz_did["e_dft_in_dft"], huz_driver._global_ks().e_tot)
    assert np.isclose(mu_did["e_dft_in_dft"], huz_did["e_dft_in_dft"])


@pytest.mark.parametrize("driver", ["mu_driver", "huz_driver"])
def test_embedded_ccsd(driver, request):
    driver = request.getfixturevalue(driver)
    # assert np.isclose(
    #     huz_driver._run_emb_fci(huz_driver.embedded_scf).e_tot, -51.61379094995273
    # )
    ccsd, ecorr = driver._run_emb_ccsd(driver.embedded_scf)
    projector_result = getattr(driver, f"{driver.config.projector.value}")
    e_emb = (
        ccsd.e_tot
        + driver.e_env
        + driver.two_e_cross
        - projector_result["correction"]
        - projector_result["beta_correction"]
    )

    assert np.isclose(e_emb, -75.1285849238916)
    assert np.isclose(ecorr, -0.00477765364464925)


@pytest.mark.parametrize("driver", ["mu_driver", "huz_driver"])
def test_embedded_fci(driver, request):
    driver = request.getfixturevalue(driver)

    # assert np.isclose(
    #     huz_driver._run_emb_fci(huz_driver.embedded_scf).e_tot, -51.61379094995273
    # )
    fci = driver._run_emb_fci(driver.embedded_scf)
    projector_result = getattr(driver, f"{driver.config.projector.value}")
    e_emb_fci = (
        fci.e_tot
        + driver.e_env
        + driver.two_e_cross
        - projector_result["correction"]
        - projector_result["beta_correction"]
    )
    assert np.isclose(e_emb_fci, -75.12858550813999)


def test_restricted_projector_results_match(mu_driver, huz_driver) -> None:
    assert mu_driver.mu is not {} and mu_driver.huzinaga is None
    assert huz_driver.huzinaga is not {} and huz_driver.mu is None
    assert mu_driver.mu.keys() == huz_driver.huzinaga.keys()


def test_unrestricted_projector_results_match(mu_driver, huz_driver) -> None:
    assert mu_driver.mu is not {} and mu_driver.huzinaga is None
    assert huz_driver.huzinaga is not {} and huz_driver.mu is None
    assert mu_driver.mu.keys() == huz_driver.huzinaga.keys()


def test_projectors_scf_match(mu_driver, huz_driver) -> None:
    mu_scf = mu_driver.embedded_scf
    huz_scf = huz_driver.embedded_scf
    assert mu_scf.converged is True
    assert huz_scf.converged is True

    assert type(mu_scf) is type(huz_scf)
    assert mu_scf.mo_coeff.shape == huz_scf.mo_coeff.shape
    assert mu_scf.mo_occ.shape == huz_scf.mo_occ.shape
    assert mu_scf.mo_energy.shape == huz_scf.mo_energy.shape
    assert np.isclose(mu_scf.e_tot, huz_scf.e_tot)


def test_unrestricted_projectors_scf_match(mu_driver, huz_driver) -> None:
    mu_scf = mu_driver.embedded_scf
    huz_scf = huz_driver.embedded_scf
    assert mu_scf.converged is True
    assert huz_scf.converged is True

    assert type(mu_scf) is type(huz_scf)
    assert mu_scf.mo_coeff.shape == huz_scf.mo_coeff.shape
    assert mu_scf.mo_occ.shape == huz_scf.mo_occ.shape
    assert mu_scf.mo_energy.shape == huz_scf.mo_energy.shape
    assert np.isclose(mu_scf.e_tot, huz_scf.e_tot)


def test_incorrect_geometry_path() -> None:
    """Test to make sure that FileNotFoundError is thrown if invalid path to xyz geometry file is given"""
    molecule = "THIS/IS/NOT/AN/XYZ/FILE"

    with pytest.raises(ValidationError):
        config = NbedConfig(
            geometry=molecule,
            n_active_atoms=1,
            basis="STO-3G",
            xc_functional="b3lyp5",
            projector="mu",
            localization="spade",
            convergence=1e-6,
            run_ccsd_emb=True,
            run_fci_emb=True,
        )
        # match will match with any printed error message


def test_driver_standard_xyz_string_input(spinless_driver) -> None:
    """test to check driver works... raw xyz string given"""

    assert isinstance(spinless_driver.embedded_scf, StreamObject)
    assert isclose(spinless_driver.classical_energy, -3.5867934952241356)
    assert spinless_driver.embedded_scf.mo_coeff.shape == (2, 7, 6)
    logger.info(spinless_driver.embedded_scf.mo_coeff)
    assert np.all(
        spinless_driver.embedded_scf.mo_occ
        == np.array([[1, 1, 1, 1, 0, 0], [1, 1, 1, 1, 0, 0]])
    )


def test_subsystem_dft(water_filepath) -> None:
    """Check thatcmponenets match total dft energy."""
    config = NbedConfig(
        geometry=str(water_filepath),
        n_active_atoms=2,
        basis="STO-3G",
        xc_functional="b3lyp",
        projector="mu",
        localization="spade",
        convergence=1e-6,
        run_ccsd_emb=False,
        run_fci_emb=False,
    )

    driver = NbedDriver(config)
    driver.embed()

    energy_DFT_components = (
        driver.e_act
        + driver.e_env
        + driver.two_e_cross
        + driver._global_ks.energy_nuc()
    )

    assert isclose(energy_DFT_components, driver._global_ks.e_tot)


if __name__ == "__main__":
    pass
